HEADER_PEPTIDE = 'Peptide sequence'
HEADER_NO_PSM = 'PSM count'
HEADER_AREA = 'MS1 area (highest of all PSMs)'
HEADER_QVAL = 'q-value'
HEADER_PEP = 'PEP'
HEADER_QVAL_MODELED = 'q-value (linear modeled)'
HEADER_FALSE_LOC_RATE = 'PTM FLR'
HEADER_LINKED_PSMS = 'Files/scans for peptide'
HEADER_BAREPEP = 'Bare peptide'
HEADER_PROTEINS = 'Protein(s)'
HEADER_MASTERPROTEINS = 'Master protein(s)'
HEADER_STARTSTOP = 'Protein start/stop site(s)'
HEADER_NO_CONTENTPROTEINS = 'Total number proteins in group'
HEADER_COVERAGES = 'Coverage(s)'
HEADER_GENES = 'Gene ID(s)'
HEADER_ASSOCIATED = 'Gene name(s)'
HEADER_DESCRIPTIONS = 'Description(s)'
HEADER_SVM_SCORE = 'percolator svm-score'



from app.lookups.sqlite.base import ResultLookupInterface


class SpectraDB(ResultLookupInterface):
    def add_tables(self, tabletypes):
        self.create_tables(['biosets', 'mzmlfiles'])
        self.create_tables(['mzml', 'ioninjtime', 'ionmob'])

    def store_biosets(self, biosets):
        self.store_many('INSERT INTO biosets(set_name) VALUES (?)', biosets)

    def store_mzmlfiles(self, mzmlfiles):
        self.store_many('INSERT INTO mzmlfiles(mzmlfilename, set_id) '
                        'VALUES (?, ?)', mzmlfiles)

    def index_biosets(self):
        self.index_column('setname_index', 'biosets', 'set_name')
        self.index_column('setid_index', 'biosets', 'set_id')
        self.index_column('setid_mzmlfn_index', 'mzmlfiles', 'set_id')
        self.index_column('mzmlfn_index', 'mzmlfiles', 'mzmlfilename')
        self.index_column('mzmlfnid_index', 'mzmlfiles', 'mzmlfile_id')

    def get_setnames(self):
        cursor = self.get_cursor()
        cursor.execute('SELECT set_name, set_id FROM biosets')
        return {setname: set_id for setname, set_id in cursor.fetchall()}
    def store_mzmls(self, spectra, ioninj, ionmob):
        self.store_many(
            'INSERT INTO mzml(spectra_id, mzmlfile_id, scan_sid, charge, mz, '
            'retention_time) '
            'VALUES (?, ?, ?, ?, ?, ?)', spectra)
        self.store_many(
            'INSERT INTO ioninjtime(spectra_id, ion_injection_time) '
            'VALUES(?, ?)', ioninj)
        self.store_many(
            'INSERT INTO ionmob(spectra_id, ion_mobility) '
            'VALUES(?, ?)', ionmob)

    def index_mzml(self):
        self.index_column('spectra_id_index', 'mzml', 'spectra_id')
        self.index_column('mzmlfnid_mzml_index', 'mzml', 'mzmlfile_id')
        self.index_column('scan_index', 'mzml', 'scan_sid')
        self.index_column('specrt_index', 'mzml', 'retention_time')
        self.index_column('specmz_index', 'mzml', 'mz')
        self.index_column('ionm_sp_ix', 'ionmob', 'spectra_id')
        self.index_column('ionin_sp_ix', 'ioninjtime', 'spectra_id')



from math import log

from app.dataformats import peptable as peptabledata
from app.dataformats import prottable as prottabledata
from app.dataformats import mzidtsv as mzidtsvdata

from app.actions.psmtopeptable import evaluate_peptide
from app.readers import fasta


def generate_bestpep_proteins(peptides, scorecol, minlog, outputaccfield, 
        protcol, higherbetter=True):
    """Best peptide for each protein in a table"""
    protein_peptides = {}
    if minlog:
        higherbetter = False
    if not protcol:
        protcol = peptabledata.HEADER_MASTERPROTEINS
    for peptide in peptides:
        p_acc = peptide[protcol]
        if ';' in p_acc or p_acc == 'NA':
            continue
        protein_peptides = evaluate_peptide(protein_peptides, peptide, p_acc,
                higherbetter, scorecol, fncol=False)
    if minlog:
        try:
            nextbestscore = min([pep['score'] for pep in
                                 protein_peptides.values()
                                 if pep['score'] > 0])
        except ValueError:
            print('WARNING: Cannot find score of type {} which is above 0. '
                    'Only scores above zero can have a -log value.'.format(scorecol))
            nextbestscore = -log(1e-06, 10) # 1 decoy in a million, fake value
        else:
            nextbestscore = -log(nextbestscore, 10)
    for protein, peptide in protein_peptides.items():
        if minlog and peptide['score'] != 'NA':
            peptide['score'] = log_score(peptide['score'], nextbestscore)
        yield {
                outputaccfield: protein, 
                prottabledata.HEADER_QSCORE: str(peptide['score'])
                }


def log_score(score, nextbestscore):
    try:
        logged_score = -log(score, 10)
    except ValueError:
        logged_score = nextbestscore + 2
    return logged_score


def generate_classic_fdr(target, decoy, featfield):
    '''Normal FDR in which we just rank all the proteins/genes and do not pick from pairs'''
    tproteins = [x for x in target if get_score(x) is not None]
    dproteins = [x for x in decoy if get_score(x) is not None]
    [x.update({'target_decoy': 'target'}) for x in tproteins]
    [x.update({'target_decoy': 'decoy'}) for x in dproteins]
    proteins = sorted(tproteins + dproteins, key=lambda x: get_score(x),
                      reverse=True)
    #fdrheader = headerfields['proteinfdr'][prottabledata.HEADER_QVAL][None]
    # FIXME complex stupid headerfields system, remove it?
    fdrheader = prottabledata.HEADER_QVAL
    for protein in qvalue_generator(fdrheader, proteins, featfield):
        yield protein


def generate_pick_fdr(target, decoy, tfastafn, dfastafn, featfield,
        fastadelim, genefield):
    t_scores, d_scores = {}, {}
    for protein in target:
        acc = protein[featfield]
        t_scores[acc] = protein
        t_scores[acc]['target_decoy'] = 'target'
    for protein in decoy:
        acc = protein[featfield]
        d_scores[acc] = protein
        d_scores[acc]['target_decoy'] = 'decoy'
    # FIXME make sure pairs are correct?
    # I mean, they are just names really - we should probably verify the T/D fasta?
    # Also use the concatenated DB instead, which is sort of the point of picking
    prefixlen = len(mzidtsvdata.DECOY_PREFIX)
    if featfield == prottabledata.HEADER_PROTEIN:
        acctype = 'protein'
    elif featfield == prottabledata.HEADER_GENENAME:
        acctype = 'genename'
    elif featfield == prottabledata.HEADER_GENEID:
        acctype = 'ensg'
    else:
        print('Error determining accession type of table to calculate FDR for')
        sys.exit(1)
    tfasta = fasta.get_genes_pickfdr(tfastafn, acctype, fastadelim, genefield)
    dfasta = fasta.get_genes_pickfdr(dfastafn, acctype, fastadelim, genefield)
    tdmap = {}
    for target, decoy in zip(tfasta, dfasta):
        tdmap[target] = decoy
    picked_proteins = []
    for tgene, dgene in tdmap.items():
        picked = pick_target_decoy(t_scores.get(tgene), d_scores.get(dgene))
        if picked:
            picked_proteins.append(picked)
    sorted_proteins = sorted(picked_proteins, key=lambda x: get_score(x),
                             reverse=True)
    #fdrheader = headerfields['proteinfdr'][prottabledata.HEADER_QVAL][None]
    fdrheader = prottabledata.HEADER_QVAL
    for protein in qvalue_generator(fdrheader, sorted_proteins, featfield):
        yield protein


def qvalue_generator(fdrheader, sorted_features, featfield):
    tdcounter = {'target': 0, 'decoy': 0}
    previousscore = get_score(sorted_features[0])
    outfeats = []
    for feat in sorted_features:
        outfeat = {k: v for k, v in feat.items()}
        score = get_score(outfeat)
        if score != previousscore:
            # new score, all proteins with previous score get same fdr
            previousscore = score
            try:
                fdr = tdcounter['decoy'] / float(tdcounter['target'])
            except ZeroDivisionError:
                fdr = 1
            for feat in sorted(outfeats, key=lambda x: x[featfield]):
                feat[fdrheader] = fdr
                yield feat
            outfeats = []
        tdcounter[outfeat['target_decoy']] += 1
        # Only report target hits so FDR=D/T
        if outfeat['target_decoy'] == 'target':
            outfeats.append(outfeat)
    # All proteins from bottom of list (no new score) get FDR as well
    try:
        fdr = tdcounter['decoy'] / float(tdcounter['target'])
    except ZeroDivisionError:
        fdr = 1
    for feat in sorted(outfeats, key=lambda x: x[featfield]):
        feat[fdrheader] = fdr
        yield feat


def get_score(protein):
    try:
        return float(protein[prottabledata.HEADER_QSCORE])
    except (TypeError, ValueError):
        return None


def pick_target_decoy(tfeature, dfeature):
    """Feed it with a target and decoy score and the protein/gene/id names,
    and this will return target/decoy type, the winning ID and the score"""
    tscore, dscore = get_score(tfeature), get_score(dfeature)
    if tscore is None and dscore is None:
        return False
    elif None in [tscore, dscore]:
        # return the non-False feature
        return [v for k, v in {tscore: tfeature, dscore: dfeature}.items()
                if k is not False][0]
    elif tscore == dscore:
        # same score or both False
        return False
    elif tscore > dscore:
        return tfeature
    elif tscore < dscore:
        return dfeature
    else:
        # in case uncaught edgecase occurs
        print('WARNING, target score {} and decoy score {} could not be '
              'compared'.format(tscore, dscore))
        return False


def generate_top_ms1_peptides(peptides, protcol):
    """Fed with a peptides generator, this returns the 3 PSMs with
    the highest precursor intensities (or areas, or whatever is
    given in the HEADER_PRECURSOR_QUANT"""
    top_ms1_peps = {}
    for peptide in peptides:
        protacc = peptide[protcol]
        precursor_amount = peptide[peptabledata.HEADER_AREA]
        if ';' in protacc or precursor_amount == 'NA':
            continue
        precursor_amount = float(precursor_amount)
        pepseq = peptide[peptabledata.HEADER_PEPTIDE]
        try:
            peptide_area = top_ms1_peps[protacc][pepseq]
        except KeyError:
            try:
                top_ms1_peps[protacc][pepseq] = precursor_amount
            except KeyError:
                top_ms1_peps[protacc] = {pepseq: precursor_amount}
        else:
            if precursor_amount > peptide_area:
                top_ms1_peps[protacc][pepseq] = precursor_amount
    return top_ms1_peps


def calculate_protein_precursor_quant(top_ms1_peps, prot_acc):
    try:
        amounts = top_ms1_peps[prot_acc].values()
    except KeyError:
        return 'NA'
    else:
        amounts = sorted([x for x in amounts if x > 0], reverse=True)[:3]
        return sum(amounts) / len(amounts)


def add_ms1_quant_from_top3_mzidtsv(features, peptides, outputaccfield, featcol):
    """Collects peptides with the highest precursor quant values,
    adds sum of the top 3 of these to a protein table"""
    top_ms1_peps = generate_top_ms1_peptides(peptides, featcol)
    for feat in features:
        acc = feat[outputaccfield]
        prec_area = calculate_protein_precursor_quant(top_ms1_peps, acc)
        outfeat = {k: v for k, v in feat.items()}
        outfeat[prottabledata.HEADER_AREA] = str(prec_area)
#        outprotein[headerfields['precursorquant'][
#            prottabledata.HEADER_AREA][None]] = str(prec_area)
        yield outfeat 



from app.drivers import base
from app.drivers.options import lookup_options, sequence_options

from app.readers import spectra as spectrareader
from app.readers import openms as openmsreader
from app.readers import tsv as tsvreader

from app.actions.lookups import spectra as spectralookup
from app.actions.lookups import quant as quantlookups 
from app.actions.lookups import sequence as seqlookups


# FIXME infn/outfn is special and should be speicifed on driver, or on baselookup?
# Or is it only used in storeseq?


class SpectraLookupDriver(base.LookupDriver):
    lookuptype = 'spectra'
    command = 'storespectra'
    commandhelp = ('Create lookup of spectra in mzML format. '
            'Requires passing mzML files to --spectra. '
            'Biological set names for each file should be specified using '
            '--setnames')

    def parse_input(self, **kwargs):
        super().parse_input(**kwargs)
        if self.setnames is None:
            assert self.lookup is not None, ('Must specify lookup '
                                             'if setnames have not '
                                             'been provided')
        else:
            self.setnames = [x.replace('"', '') for x in self.setnames]

    def set_options(self):
        super().set_options()
        self.options['lookupfn'].update({'required': False, 'default': None})
        self.options.update(self.define_options(['outfile', 'spectrafns', 'setnames'],
                                                lookup_options))

    def create_lookup(self):
        spectralookup.create_bioset_lookup(self.lookup, self.spectrafns,
                                          self.setnames)
        fn_spectra = spectrareader.mzmlfn_ms2_spectra_generator(
            self.spectrafns)
        spectralookup.create_spectra_lookup(self.lookup, fn_spectra)


class QuantLookupDriver(base.LookupDriver):
    command = 'storequant'
    lookuptype = 'specquant'

    commandhelp = ('Create lookup of isobaric quant data in OpenMS ' # FIXME
                   'consensusXML format. Use requires matched files passed to '
                   '--spectra, --ms1, --isobaric parameters, and --dbfile '
                   'with an sqlite lookup of already stored spectra.'
                   'Creates lookup of precursor quant data in OpenMS '
                   'featureXML, or Kronik output formats. '
                   'Use requires --spectra, --dbfile with an sqlite lookup of '
                   'spectra, --quanttype to determine quant output, --mztol, '
                   '--rttol, --mztoltype for tolerance specification, and '
                   'passing a featureXML or kronik file to -i.')



    def set_options(self):
        super().set_options()
        self.options.update(self.define_options([
            'spectrafns', 'kronik', 'dinosaur', 'isobaric',
            'sum_or_apex', 'rttol', 'mztol', 'mztoltype'], lookup_options))

    def parse_input(self, **kwargs):
        super().parse_input(**kwargs)
        if getattr(self, 'isobaricfns') and len(self.isobaricfns) > 0:
            self.tabletypes.append('isobaric')
        else:
            self.isobaricfns = False
        self.ms1type = False
        if getattr(self, 'dinosaurfns') and len(self.dinosaurfns) > 0:
            self.tabletypes.append('ms1')
            self.ms1type = 'dinosaur'
            self.ms1fns = self.dinosaurfns[:]
        elif getattr(self, 'kronikfns') and len(self.kronikfns) > 0:
            self.tabletypes.append('ms1')
            self.ms1type = 'kronik'
            self.ms1fns = self.kronikfns[:]

    def create_lookup(self):
        if self.isobaricfns:
            quantlookups.create_isobaric_quant_lookup(self.lookup, self.spectrafns, self.isobaricfns)
        if self.ms1type:
            ms1feats = tsvreader.mzmlfn_tsvfeature_generator(self.spectrafns, self.ms1fns)
            quantlookups.create_precursor_quant_lookup(self.lookup, ms1feats,
                    self.sum_or_apex, self.ms1type, self.rt_tol, self.mz_tol, self.mz_toltype)


class SequenceLookupDriver(base.LookupDriver):
    """Creates an SQLite lookup DB from a FASTA file. Sequences are
    trypsinized and stored. It's possible to store sequences reversed
    for N-terminal insource fragmentation indexing, and it can be specified to cut
    tryptic before proline.
    """
    lookuptype = 'searchspace'
    command = 'storeseq'
    commandhelp = """Create a lookup DB from a FASTA file. Sequences are
    either stored as 1-aa-shifted overlapping peptides for full-length-protein 
    matching, or as trypsinized peptides."""

    def __init__(self):
        super().__init__()
        self.infiletype = 'FASTA'

    def set_options(self):
        super().set_options()
        self.options.update(self.define_options(['fn', 'outfile']))
        self.options['lookupfn'].update({'required': False, 'default': None})
        self.options.update(self.define_options(['fullprotein', 'pepprotmapping', 'insourcefrag',
            'ntermmloss', 'proline', 'minlength'], lookup_options))
        self.options.update(self.define_options(['trypsinize', 'miss_cleavage'],
            sequence_options))

    def create_lookup(self):
        if self.fullprotein:
            if not self.minlength:
                self.minlength = 7
                # Default here since it is a param shared with many others 
                # where it can be 0. TODO create own param for this driver
                print('--minlen not specified, default to 7')
            print('Creating full-length protein lookup, minimum matching '
                    'peptide length is {}'.format(self.minlength))
            if self.proline or self.insourcefrag or not self.trypsinize or self.miss_cleavage:
                print('Ignoring other options for tryptic lookup building')
            seqlookups.create_searchspace_wholeproteins(self.lookup, self.fn,
                                                         self.minlength)
        elif self.pepprotmapping:
            seqlookups.create_searchspace_map_accessions(self.lookup, self.fn, self.minlength, self.proline,
                    self.insourcefrag, self.trypsinize, self.miss_cleavage, self.ntermmloss)
        else:
            seqlookups.create_searchspace(self.lookup, self.fn, self.minlength, self.proline,
                    self.insourcefrag, self.trypsinize, self.miss_cleavage, self.ntermmloss)



import re
from itertools import product

from app.readers import percolator as reader
from app.readers import xmlformatting as formatting


def filter_whole_proteins(elements, lookup, seqtype, ns, deamidation, minpeplen, enforce_tryp):
    # Remove duplicate sequences by using them as keys
    for element in elements:
        seq_matches_protein = False
        element_seqs = get_seqs_from_element(element, seqtype, ns, deamidation)
        element_prots = lookup.get_proteins_from_peps(element_seqs, minpeplen)
        for pepseq, proteins in element_prots.items():
            for prot_id, pos, protseq in proteins:
                if pepseq in protseq.replace('L', 'I'):
                    if enforce_tryp and (pos == 0 or not set(
                            [pepseq[-1],
                             protseq[pos - 1]]).difference(['K', 'R'])):
                        # pepseq is tryptic on both ends, or
                        # pepseq is an N-term peptide),
                        # matches to protein seq so remove
                        seq_matches_protein = True
                        break
                    elif not enforce_tryp:
                        seq_matches_protein = True
                        break
        if seq_matches_protein:
            formatting.clear_el(element)
        else:
            yield formatting.string_and_clear(element, ns)


def dedup_psms(psms, ns):
    psm_idseqs = set()
    for psm in psms:
        psm_id = psm.attrib['{%s}psm_id' % ns['xmlns']]
        psm_seq = reader.get_psm_seq(psm, ns).upper()
        psm_idseq = f'{psm_id}_{psm_seq}'
        if psm_idseq not in psm_idseqs:
            psm_idseqs.add(psm_idseq)
            yield formatting.string_and_clear(psm, ns)


def passthrough(elements, ns):
    for el in elements:
        yield formatting.string_and_clear(el, ns)


def filter_best_peptide(peptide_elements, ns):
    '''Treats percolator XML that for some reason has changed and gotten duplicate
    peptides (e.g. when the UNIMODs have been removed. PSMs will be returned as-is, but
    Peptides will be uniquified (best scoring peptide will be retained)
    '''
    peptide_scores = {}
    for pep in peptide_elements:
        seq = reader.get_peptide_seq(pep, ns).upper()
        score = float(pep.find('{%s}svm_score' % ns['xmlns']).text)
        if seq not in peptide_scores or score > peptide_scores[seq][0]:
            peptide_scores[seq] = (score, formatting.string_and_clear(pep, ns))
    for _score, pepstring in peptide_scores.values():
        yield pepstring


def filter_known_searchspace(elements, seqtype, lookup, ns, ntermwildcards,
                             deamidation):
    """Yields peptides from generator as long as their sequence is not found in
    known search space dict. Useful for excluding peptides that are found in
    e.g. ENSEMBL or similar"""
    for element in elements:
        seq_is_known = False
        for seq in get_seqs_from_element(element, seqtype, ns, deamidation):
            if lookup.check_seq_exists(seq, ntermwildcards):
                seq_is_known = True
                break
        if seq_is_known:
            formatting.clear_el(element)
        else:
            yield formatting.string_and_clear(element, ns)


def sequence_to_filterseqs(seq, deamidation):
    # Exchange leucines for isoleucines since MS can't differ and we
    # don't want to find 'novel' peptides which only have a difference
    # in this amino acid
    seq = seq.replace('L', 'I')
    if deamidation:
        options = [(c,) if c != 'D' else ('D', 'N') for c in seq]
        return list(''.join(o) for o in product(*options))
    else:
        return [seq]


def get_seqs_from_element(element, seqtype, ns, deamidation):
    seq = {'psm': reader.get_psm_seq, 'pep': reader.get_peptide_seq}[seqtype](element, ns).upper()
    seq = re.sub('\[UNIMOD:\d*\]', '', seq)
    return sequence_to_filterseqs(seq, deamidation)



import sys

from app.drivers.base import PepProttableDriver

from app.drivers.options import prottable_options
from app.readers import tsv as tsvreader
from app.dataformats import prottable as prottabledata
from app.dataformats import peptable as peptabledata
from app.dataformats import mzidtsv as mzidtsvdata

from app.actions import proteins
from app.actions.psmtable import isosummarize


class ProttableDriver(PepProttableDriver):

    def set_options(self):
        super().set_options()
        options = self.define_options(['decoyfn', 'scorecolpattern', 'minlogscore',
            'quantcolpattern', 'minint', 'denomcols', 'denompatterns', 'mediansweep',
            'medianintensity', 'median_or_avg', 'logisoquant', 'mediannormalize',
            'mednorm_factors', 'keep_psms_na', 'precursor', 'psmfile', 'fastadelim', 
            'genefield', 't_fasta', 'd_fasta', 'fdrtype'], prottable_options)
        self.options.update(options)

    def get_td_proteins_bestpep(self, theader, dheader):
        self.header = [self.headeraccfield] + prottabledata.PICKED_HEADER
        tscorecol = tsvreader.get_cols_in_file(self.scorecolpattern, theader, True)
        dscorecol = tsvreader.get_cols_in_file(self.scorecolpattern, dheader, True)
        tpeps = tsvreader.generate_split_tsv_lines(self.fn, theader)
        dpeps = tsvreader.generate_split_tsv_lines(self.decoyfn, dheader)
        targets = proteins.generate_bestpep_proteins(tpeps, tscorecol, 
                self.minlogscore, self.headeraccfield, self.fixedfeatcol)
        decoys = proteins.generate_bestpep_proteins(dpeps, dscorecol,
                self.minlogscore, self.headeraccfield, self.fixedfeatcol)
        return targets, decoys

    def get_quant(self, theader, features):
        if self.precursor:
            tpeps = tsvreader.generate_split_tsv_lines(self.fn, theader)
            self.header.append(prottabledata.HEADER_AREA)
            features = proteins.add_ms1_quant_from_top3_mzidtsv(features, 
                    tpeps, self.headeraccfield, self.fixedfeatcol)
        if self.quantcolpattern:
            psmheader = tsvreader.get_tsv_header(self.psmfile)
            denomcols = False
            if self.denomcols is not None:
                denomcols = [self.number_to_headerfield(col, psmheader)
                             for col in self.denomcols]
            elif self.denompatterns is not None:
                denomcolnrs = [tsvreader.get_columns_by_combined_patterns(psmheader,
                    [self.quantcolpattern, pattern]) for pattern in self.denompatterns]
                denomcols = set([col for cols in denomcolnrs for col in cols])
            elif not self.mediansweep and not self.medianintensity:
                print('Must define either denominator column numbers '
                        'or regex pattterns to find them, or use median sweep, or '
                        'report median intensities.')
                sys.exit(1)
            elif self.medianintensity and self.mediannormalize:
                print('Cannot do median-centering on intensity values, exiting')
                sys.exit(1)
            quantcols = tsvreader.get_columns_by_pattern(psmheader, self.quantcolpattern)
            mn_factors = False
            if self.mednorm_factors:
                mnhead = tsvreader.get_tsv_header(self.mednorm_factors)
                mn_factors = tsvreader.generate_split_tsv_lines(self.mednorm_factors, mnhead)
            nopsms = [isosummarize.get_no_psms_field(qf) for qf in quantcols]
            self.header = self.header + quantcols + nopsms + [prottabledata.HEADER_NO_FULLQ_PSMS]
            features = isosummarize.get_isobaric_ratios(self.psmfile, psmheader,
                    quantcols, denomcols, self.mediansweep, self.medianintensity,
                    self.median_or_avg, self.minint, features, self.headeraccfield,
                    self.fixedfeatcol, False, False, False, self.logisoquant, self.mediannormalize,
                    mn_factors, self.keepnapsms)
        return features

    def set_features(self):
        theader = tsvreader.get_tsv_header(self.fn)
        dheader = tsvreader.get_tsv_header(self.decoyfn)
        targets, decoys = self.get_td_proteins_bestpep(theader, dheader)
        if self.fdrtype == 'picked':
            if not self.t_fasta or not self.d_fasta:
                print('Must use --targetfasta and --decoyfasta when using picked FDR')
                sys.exit(1)
            fastadelim, genefield = self.get_fastadelim_genefield(self.fastadelim,
                    self.genefield)
            features = proteins.generate_pick_fdr(
               targets, decoys, self.t_fasta, self.d_fasta, self.headeraccfield,
               fastadelim, genefield)
        else:
            features = proteins.generate_classic_fdr(targets, decoys, self.headeraccfield)
        self.features = self.get_quant(theader, features)


class ProteinsDriver(ProttableDriver):
    command = 'proteins'
    commandhelp = 'Create a protein table from peptides'
    outsuffix = '_proteins.tsv'
    headeraccfield = prottabledata.HEADER_PROTEIN
    fixedfeatcol = peptabledata.HEADER_MASTERPROTEINS


class GenesDriver(ProttableDriver):
    command = 'genes'
    commandhelp = 'Create a gene table from peptides'
    outsuffix = '_genes.tsv'
    headeraccfield = prottabledata.HEADER_GENENAME
    fixedfeatcol = mzidtsvdata.HEADER_SYMBOL


class ENSGDriver(GenesDriver):
    command = 'ensg'
    commandhelp = 'Create an ENSG table from peptides'
    outsuffix = '_ensg.tsv'
    headeraccfield = prottabledata.HEADER_GENEID
    fixedfeatcol = mzidtsvdata.HEADER_GENE



from app.lookups.sqlite.base import DatabaseConnection


class SearchSpaceDB(DatabaseConnection):
    def add_tables(self, tabletypes):
        """Creates a searchspace lookup sqlite."""
        self.create_tables(['known_searchspace', 'protein_peptides', 'proteins', 'protein_seq'])

    def write_peps(self, peps):
        """Writes peps to db. We can reverse to be able to look up
        peptides that have some amino acids missing at the N-terminal.
        This way we can still use the index.
        """
        cursor = self.get_cursor()
        cursor.executemany(
            'INSERT OR IGNORE INTO known_searchspace(seqs) VALUES (?)', peps)
        self.conn.commit()

    def index_peps(self, reverse_seqs):
        if reverse_seqs:
            self.index_column('reverse_seqs_index', 'known_searchspace',
                              'seqs COLLATE NOCASE', unique=True)
        else:
            self.index_column('sequence_index', 'known_searchspace',
                              'seqs', unique=True)

    def store_pep_proteins(self, pepproteins):
        cursor = self.get_cursor()
        cursor.executemany('INSERT INTO proteins(protein_acc) VALUES(?)', ((x,) for x in pepproteins))
        cursor.executemany('INSERT INTO protein_seq(protein_acc, sequence) VALUES(?,?)',
                ((k, v['seq']) for k, v in pepproteins.items()))
        cursor.executemany('INSERT INTO protein_peptides(seq, protein_acc, pos) '
                           'VALUES(?, ?, ?)', ((pep[0], k, pep[1]) for k,v in pepproteins.items()
                               for pep in v['peps']))
        self.conn.commit()

    def store_tryp_peps_mapped(self, pepproteins):
        """Writes peps to db, but this time peps is a dict where {prot1: [PEPSEQ, IAMAPEPTIDE, ...], ...},
        and we also store the mapping to a peptide's proteins. We can reverse to be able to look up
        peptides that have some amino acids missing at the N-terminal.
        This way we can still use the index.
        """
        cursor = self.get_cursor()
        cursor.executemany('INSERT INTO proteins(protein_acc) VALUES(?)', ((x,) for x in pepproteins))
        cursor.executemany('INSERT INTO protein_peptides(seq, protein_acc, pos) VALUES(?, ?, ?)',
                ((pep, prot, 0) for prot,peps in pepproteins.items() for pep in peps))

    def index_proteins(self):
        self.index_column('pepix', 'protein_peptides', 'seq')
        self.index_column('pp_protix', 'protein_peptides', 'protein_acc')
        self.index_column('pseq_protix', 'protein_seq', 'protein_acc')
        self.conn.commit()

    def index_peps_mapped(self, reverse_seqs):
        collate = ' COLLATE NOCASE' if reverse_seqs else ''
        self.index_column('reverse_seqs_index', 'protein_peptides', f'seq{collate}', unique=True)
        self.index_column('pp_protix', 'protein_peptides', 'protein_acc')
        self.conn.commit()

    def get_multi_seq(self, allseqs):
        cursor = self.get_cursor()
        maxparam = 999
        allseqs_found = set()
        for i in range(0, len(allseqs), maxparam):
            seqs = allseqs[i:i+maxparam]
            sql = 'SELECT seqs FROM known_searchspace WHERE seqs IN ({})'.format(
                    ', '.join('?' for _ in seqs))
            [allseqs_found.add(x[0]) for x in cursor.execute(sql, seqs)]
        return allseqs_found

    def check_seq_exists(self, seq, amount_ntermwildcards):
        """Look up sequence in sqlite DB. Returns True or False if it
        exists (or not). When looking up a reversed DB with
        ntermwildcards: we reverse the sequence of the pep and add
        a LIKE and %-suffix to the query.
        """
        cursor = self.get_cursor()
        if amount_ntermwildcards > 0:
            seq = seq[::-1]
            sqlseq = '{}{}'.format(seq, '%')
            # FIXME non-parametrized string binding because if ? binding is
            # used the INDEX is not used when looking up, apparently because
            # the query cant be optimized when using LIKE and binding.
            sql = ('select seqs from known_searchspace where seqs LIKE '
                   '"{}"'.format(sqlseq))
            for match in cursor.execute(sql):
                if match[0][:-amount_ntermwildcards] in seq:
                    return True
            return False
        else:
            sql = ('select exists(select seqs from known_searchspace '
                   'where seqs=? limit 1)')
            return cursor.execute(sql, (seq, )).fetchone()[0] == 1

    def get_proteins_from_peps(self, peptides, minpeplen):
        '''Retrieve peptide/proteinseq combinations'''
        cursor = self.get_cursor()
        sql = ('SELECT pp.protein_acc, pp.pos, ps.sequence FROM protein_peptides AS pp '
            'INNER JOIN protein_seq AS ps ON ps.protein_acc=pp.protein_acc WHERE pp.seq=?')
        return {seq: [(protid, pos, pseq) for protid, pos, pseq in
            cursor.execute(sql, (seq[:minpeplen],))] for seq in peptides}

    def get_proteins_matching_peptides(self, peptides, amount_ntermwildcards, minlen):
        '''Retrieve peptide/proteinseq combinations'''
        cursor = self.get_cursor()
        if amount_ntermwildcards > 0:
            protids = set()
            for seq in peptides:
                findseq = f'{seq[::-1]}'
                sql = (f'SELECT pp.seq, pp.protein_acc FROM protein_peptides AS pp WHERE pp.seq LIKE "{findseq}%"')
                for match in cursor.execute(sql):
                    if match[0][:-amount_ntermwildcards] in findseq:
                        protids.add(match[1])

        elif minlen:
            sql = ('SELECT pp.protein_acc, pp.pos, ps.sequence FROM protein_peptides AS pp '
            'INNER JOIN protein_seq AS ps ON ps.protein_acc=pp.protein_acc WHERE pp.seq=?')
            protids = {seq: [(protid, pos, pseq) for protid, pos, pseq in 
                    cursor.execute(sql, (seq[:minlen],))] for seq in peptides}
        else:
            sql = ('SELECT pp.protein_acc FROM protein_peptides AS pp WHERE pp.seq=?')
            protids = {protid[0] for seq in peptides for protid in cursor.execute(sql, (seq,))}
        return protids



#!/usr/bin/python3

import sys
from app.drivers import (startup, lookups, sequence, psmtable, percolator,
        peptable, prottable, merge)


def main():
    drivers = [lookups.SpectraLookupDriver(),
               lookups.QuantLookupDriver(),
               lookups.SequenceLookupDriver(),

               sequence.DecoySeqDriver(),
               sequence.TrypsinizeDriver(),

               psmtable.TSVConcatenateDriver(),
               psmtable.TSVSplitDriver(),
               psmtable.Perco2PSMDriver(),
               psmtable.ConfidenceFilterDriver(),
               psmtable.PSMTableRefineDriver(),
               psmtable.IsoSummarizeDriver(),
               psmtable.DeleteSetDriver(),
               psmtable.DuplicateFilterDriver(),
               psmtable.SequenceFilterDriver(),
               psmtable.SequenceMatchDriver(),

               percolator.FilterSequences(),
               percolator.FilterDuplicatePeptideDriver(),
               percolator.SplitProteinDriver(),

               peptable.CreatePeptableDriver(),
               prottable.ProteinsDriver(),
               prottable.GenesDriver(),
               prottable.ENSGDriver(),
               
               merge.MergeDriver(),
               ]
    startup.start_msstitch(drivers, sys.argv)



import os

DB_STORE_CHUNK = 500000


def create_bioset_lookup(lookupdb, spectrafns, set_names):
    """Fills lookup database with biological set names"""
    unique_setnames = set(set_names)
    lookupdb.store_biosets(((x,) for x in unique_setnames))
    set_id_map = lookupdb.get_setnames()
    mzmlfiles = ((os.path.basename(fn), set_id_map[setname])
                 for fn, setname in zip(spectrafns, set_names))
    lookupdb.store_mzmlfiles(mzmlfiles)
    lookupdb.index_biosets()


def create_spectra_lookup(lookup, fn_spectra):
    """Stores all spectra rt, injection time, and scan ID in db"""
    dbspectra, dbioninj, dbionmob = [], [], []
    count = 0
    mzmlmap = lookup.get_mzmlfile_map()
    for fn, spectrum in fn_spectra:
        spec_id = '{}_{}'.format(mzmlmap[fn], spectrum['specscanid'])
        mzml_rt = round(spectrum['rt'], 12)
        mzml_iit = round(float(spectrum['iit']), 12)
        mzml_mob = round(float(spectrum['ionmob']), 12)
        mz = float(spectrum['mz'])
        dbspectra.append((spec_id, mzmlmap[fn], spectrum['specscanid'],
                         spectrum['charge'], mz, mzml_rt))
        dbioninj.append((spec_id, mzml_iit)) if mzml_iit else False
        dbionmob.append((spec_id, mzml_mob)) if mzml_mob else False
        count += 1
        if len(dbspectra) == DB_STORE_CHUNK:
            lookup.store_mzmls(dbspectra, dbioninj, dbionmob)
            dbspectra, dbioninj, dbionmob  = [], [], []
    lookup.store_mzmls(dbspectra, dbioninj, dbionmob)
    lookup.index_mzml()
    print('Stored {} spectra in lookup'.format(count))



import os
from numpy import polyfit
from math import log

from tests.integration import basetests


class TestPSM2Peptable(basetests.BaseTest):
    command = 'peptides'
    suffix = '_peptable.tsv'
    infilename = 'set1_target_pg.tsv'
    channels = ['tmt10plex_{}'.format(x) for x in ['126', '127N', '127C',
                                               '128N', '128C', '129N',
                                               '129C', '130N', '130C',
                                               '131']]
    nopsms = ['{} - Quanted PSM count'.format(ch) for ch in channels]

    def check(self, fncol):
        psms = {}
        for psm in self.tsv_generator(self.infile[0]):
            score = str(float(psm['percolator svm-score']))
            try:
                psms[psm['Peptide']][score] = psm
            except KeyError:
                psms[psm['Peptide']] = {score: psm}
        wildcardstrip = 'tmt10plex_'
        striplist = ['#SpecFile', 'Files/scans for peptide', 'MS1 area', 
                'MS1 area (highest of all PSMs)', 'q-value (linear modeled)',
                'Fully quanted PSM count']
        for peptide in self.tsv_generator(self.resultfn):
            for newkey, oldkey in zip(['Peptide', 'peptide q-value', 'Protein'],
                                      ['Peptide sequence', 'q-value', 'Protein(s)']):
                peptide[newkey] = peptide.pop(oldkey)
            mapping_psms = psms[peptide['Peptide']]
            toppsm = mapping_psms[str(max([float(score) for score
                                           in mapping_psms.keys()]))]
            testpeptide = {k: v for k, v in peptide.items()
                           if k not in striplist and not k.startswith(wildcardstrip)}
            testpsm = {k: v for k, v in toppsm.items()
                       if k not in striplist and not k.startswith(wildcardstrip)}
            self.assertEqual(testpeptide, testpsm)
            ms1s = [self.get_float_or_na(psm['MS1 area'])
                    for psm in mapping_psms.values()]
            try:
                topms1 = max(ms1s)
            except TypeError:
                # cannot do max on mix of str and float
                topms1 = max([0 if x == 'NA' else x for x in ms1s])
            self.assertEqual(str(topms1),
                             peptide['MS1 area (highest of all PSMs)'])
            psmfnscans = set(peptide['Files/scans for peptide'].split('; '))
            with open(self.infile[0]) as fp:
                fnfield = next(fp).strip().split('\t')[fncol - 1]
            self.assertEqual(psmfnscans,
                             set(['{}_{}'.format(psm[fnfield], psm['SpecID'])
                                  for psm in mapping_psms.values()]))
        self.check_modelqvals()

    def check_modelqvals(self):
        scores, qvalues = [], []
        score, fdr, qthres = 'percolator svm-score', '^q-value', 1e-5
        for line in self.tsv_generator(self.resultfn):
            if float(line[fdr[1:]]) > qthres:
                scores.append(float(line[score]))
                qvalues.append(log(float(line[fdr[1:]]), 10))
        slope, intercept = polyfit(scores, qvalues, deg=1)
        for line in self.tsv_generator(self.resultfn):
            self.assertEqual(float(line[fdr[1:] + ' (linear modeled)']),
                                   10 ** (float(line[score]) *
                                          slope + intercept))

    def check_modelqvals_none_included(self):
        for line in self.tsv_generator(self.resultfn):
            self.assertEqual(line[fdr[1:] + ' (linear modeled)'], 'NA')

    def test_psm2peptable_normalized(self):
        options = ['--spectracol', '1', '--isobquantcolpattern',
                   'tmt10plex', '--scorecolpattern', 'svm',
                   '--ms1quantcolpattern', 'MS1', 
                   '--modelqvals', '--qvalthreshold', '1e-5',
                   '--denompatterns', '126', '--median-normalize',
                   ]
        self.run_command(options)
        self.check(1)
        self.isoquant_check(os.path.join(self.fixdir, 'target_pep_quant_norm.tsv'),
                'Peptide sequence', self.channels, self.nopsms)

    def test_psm2peptable(self):
        fncol = 1
        options = ['--spectracol', str(fncol), '--isobquantcolpattern',
                   'tmt10plex', '--scorecolpattern', 'svm',
                   '--denompatterns', '126',
                   '--ms1quantcolpattern', 'MS1', 
                   '--modelqvals', '--qvalthreshold', '1e-5',
                   '--minpepnr', str(4),
                   ]
        self.run_command(options)
        self.check(fncol)
        self.isoquant_check(os.path.join(self.fixdir, 'target_peptides.tsv'),
                'Peptide sequence', self.channels, self.nopsms)

    def test_average_summarizing(self):
        options = ['--isobquantcolpattern', 'tmt10plex',
                '--scorecolpattern', 'svm',
                   '--denompatterns', '126',
                   '--summarize-average',
                   '--ms1quantcolpattern', 'MS1', 
                   '--modelqvals', '--qvalthreshold', '1e-5',
                   '--minpepnr', str(4),
                   ]
        self.run_command(options)
        self.check(1)
        self.isoquant_check(os.path.join(self.fixdir, 'target_peptides_avg.tsv'),
                'Peptide sequence', self.channels, self.nopsms)

    def test_no_spectracol_mediansweep_keep_na(self):
        options = ['--isobquantcolpattern', 'tmt10plex',
                '--scorecolpattern', 'svm',
                   '--mediansweep',
                   '--keep-psms-na-quant',
                   '--ms1quantcolpattern', 'MS1',
                   '--modelqvals', '--qvalthreshold', '1e-5',
                   '--minpepnr', str(4),
                   ]
        self.run_command(options)
        self.check(1)
        self.isoquant_check(os.path.join(self.fixdir, 'target_peptides_sweep.tsv'),
                'Peptide sequence', self.channels, self.nopsms)

    def test_totalproteome(self):
        options = ['--spectracol', '1', '--isobquantcolpattern',
                   'tmt10plex', '--scorecolpattern', 'svm',
                   '--ms1quantcolpattern', 'MS1',
                   '--modelqvals', '--qvalthreshold', '1e-5',
                   '--minpepnr', str(4),
                   '--denompatterns', '126',
                   '--totalproteome', os.path.join(self.fixdir, 'proteins.txt'),
                   ]
        self.run_command(options)
        self.check(1)
        self.isoquant_check(os.path.join(self.fixdir, 'target_peptides_totalprotnorm_nolog.txt'),
                'Peptide sequence', self.channels, self.nopsms)

    def test_totalproteome_logiso(self):
        options = ['--spectracol', '1', '--isobquantcolpattern',
                   'tmt10plex', '--scorecolpattern', 'svm',
                   '--ms1quantcolpattern', 'MS1', 
                   '--modelqvals', '--qvalthreshold', '1e-5',
                   '--minpepnr', str(4),
                   '--denompatterns', '126',
                   '--logisoquant',
                   '--totalproteome', os.path.join(self.fixdir, 'proteins_isonorm_log.txt'),
                   ]
        self.run_command(options)
        self.check(1)
        self.isoquant_check(os.path.join(self.fixdir, 'target_peptides_totalprotnorm.txt'),
                'Peptide sequence', self.channels, self.nopsms)

    def test_totalproteome_logiso_sepnorm(self):
        '''Most proteins/peptides have identical quant in this small exp. So output of
        totalprotnormalizing with median centering is pep - prot - median = 0 - median
        '''
        options = ['--spectracol', '1', '--isobquantcolpattern',
                   'tmt10plex', '--scorecolpattern', 'svm',
                   '--ms1quantcolpattern', 'MS1', 
                   '--modelqvals', '--qvalthreshold', '1e-5',
                   '--minpepnr', str(4),
                   '--denompatterns', '126',
                   '--logisoquant',
                   '--totalproteome', os.path.join(self.fixdir, 'proteins_nonorm_log.txt'),
                   '--median-normalize',
                   '--normalization-factors-table', os.path.join(self.fixdir, 'proteins_nonorm_log.txt'),
                   ]
        self.run_command(options)
        self.check(1)
        self.isoquant_check(os.path.join(self.fixdir, 'target_peptides_totalp_sepnorm.txt'),
                'Peptide sequence', self.channels, self.nopsms)



class TestProteinTable(basetests.ProttableTest):
    command = 'proteins'
    infilename = 'target_peptides.tsv'

    def test_denoms(self):
        self.specialoptions = ['--fdrtype', 'picked',
                '--targetfasta', os.path.join(self.basefixdir, 'ens99_small.fasta'),
                '--decoyfasta', os.path.join(self.fixdir, 'protrev_ens99_small.fasta'),
                ]
        self.dotest_proteintable('^q-value', 'Master protein(s)', 'Protein ID')
        expectedfn = os.path.join(self.fixdir, 'proteins_denoms_pickfdr.txt')
        self.check_lines(expectedfn, self.resultfn)

    def test_sweep(self):
        self.specialoptions = ['--keep-psms-na-quant']
        self.dotest_proteintable('^q-value', 'Master protein(s)', 'Protein ID', summarize_method='sweep')
        expectedfn = os.path.join(self.fixdir, 'proteins_sweep.txt')
        self.check_lines(expectedfn, self.resultfn)

    def test_average_summarizing(self):
        self.specialoptions = ['--summarize-average']
        self.dotest_proteintable('^q-value', 'Master protein(s)', 'Protein ID')
        expectedfn = os.path.join(self.fixdir, 'proteins_avg.txt')
        self.check_lines(expectedfn, self.resultfn)

    def test_isonormalize_log(self):
        self.specialoptions = ['--logisoquant', '--median-normalize', '--minint', '0.1']
        self.dotest_proteintable('^q-value', 'Master protein(s)', 'Protein ID')
        expectedfn = os.path.join(self.fixdir, 'proteins_isonorm_log.txt')
        self.check_lines(expectedfn, self.resultfn)

    def test_isonormalize_nolog_sweep(self):
        self.specialoptions = ['--median-normalize']
        self.dotest_proteintable('^q-value', 'Master protein(s)', 'Protein ID', summarize_method='sweep')
        expectedfn = os.path.join(self.fixdir, 'proteins_isonorm_nolog.txt')
        self.check_lines(expectedfn, self.resultfn)

    def test_no_denom_but_intensity(self):
        self.specialoptions = []
        self.dotest_proteintable('^q-value', 'Master protein(s)', 'Protein ID', summarize_method='intensity')
        expectedfn = os.path.join(self.fixdir, 'proteins_intensities.txt')
        self.check_lines(expectedfn, self.resultfn)

    def test_intensity_normalize(self):
        self.specialoptions = ['--median-normalize']
        res = self.dotest_proteintable('^q-value', 'Master protein(s)', 'Protein ID', summarize_method='intensity', should_error=True)
        if res.returncode != 0:
            self.assertEqual(res.stdout.strip(), 
                    'Cannot do median-centering on intensity values, exiting')
        else:
            self.fail('This test should error due to an invalid combination of '
                    '--median-normalize and --medianintensity')


class TestGenenameTable(basetests.ProttableTest):
    command = 'genes'
    infilename = 'target_peptides.tsv'

    def test(self):
        self.specialoptions = [
                '--fdrtype', 'picked',
                '--targetfasta', os.path.join(self.basefixdir, 'ens99_small.fasta'),
                '--decoyfasta', os.path.join(self.fixdir, 'protrev_ens99_small.fasta')]
        self.dotest_proteintable('^q-value', 'Gene Name', 'Gene Name')
        expectedfn = os.path.join(self.fixdir, 'genenames.txt')
        self.check_lines(expectedfn, self.resultfn)


class TestENSGTable(basetests.ProttableTest):
    command = 'ensg'
    infilename = 'target_peptides.tsv'

    def test(self):
        self.specialoptions = [
                '--fdrtype', 'picked',
                '--targetfasta', os.path.join(self.basefixdir, 'ens99_small.fasta'),
                '--decoyfasta', os.path.join(self.fixdir, 'protrev_ens99_small.fasta')]
        self.dotest_proteintable('svm', 'Gene ID', 'Gene ID')
        expectedfn = os.path.join(self.fixdir, 'ensg.txt')
        self.check_lines(expectedfn, self.resultfn)



DECOY_PREFIX = 'decoy_'

HEADER_SPECFILE = '#SpecFile'
HEADER_SCANNR = 'ScanNum'
HEADER_SPECSCANID = 'SpecID'
HEADER_CHARGE = 'Charge'
HEADER_PEPTIDE = 'Peptide'
HEADER_PROTEIN = 'Protein'
HEADER_GENE = 'Gene ID'
HEADER_SYMBOL = 'Gene Name'
HEADER_DESCRIPTION = 'Description'
HEADER_PRECURSOR_MZ = 'Precursor'
HEADER_PRECURSOR_QUANT = 'MS1 area'
HEADER_PREC_PURITY = 'Precursor ion fraction'
HEADER_PRECURSOR_FWHM = 'FWHM'
HEADER_MASTER_PROT = 'Master protein(s)'
HEADER_PG_CONTENT = 'Protein group(s) content'
HEADER_PG_AMOUNT_PROTEIN_HITS = 'Total number of matching proteins in group(s)'
HEADER_PG = [HEADER_MASTER_PROT, HEADER_PG_CONTENT,
             HEADER_PG_AMOUNT_PROTEIN_HITS]
HEADER_SVMSCORE = 'percolator svm-score'
HEADER_MISSED_CLEAVAGE = 'missed_cleavage'
HEADER_PSMQ = 'PSM q-value'
HEADER_PSM_PEP = 'PSM PEP'
HEADER_PEPTIDE_Q = 'peptide q-value'
HEADER_PEPTIDE_PEP = 'peptide PEP'
HEADER_TARGETDECOY = 'TD'
HEADER_MSGFSCORE = 'MSGFScore'
HEADER_EVALUE = 'EValue'
HEADER_SETNAME = 'Biological set'
HEADER_RETENTION_TIME = 'Retention time(min)'
HEADER_INJECTION_TIME = 'Ion injection time(ms)'
HEADER_ION_MOB = 'Ion mobility(Vs/cm2)'
MOREDATA_HEADER = [HEADER_RETENTION_TIME, HEADER_INJECTION_TIME, HEADER_ION_MOB]
PERCO_HEADER = [HEADER_SVMSCORE, HEADER_PSMQ, HEADER_PSM_PEP, HEADER_PEPTIDE_Q, HEADER_PEPTIDE_PEP, HEADER_TARGETDECOY]

AA_WEIGHTS_MONO = { # From ExPASY
        'A': 71.03711,
        'R': 156.10111,
        'N': 114.04293,
        'D': 115.02694,
        'C': 103.00919,
        'E': 129.04259,
        'Q': 128.05858,
        'G': 57.02146,
        'H': 137.05891,
        'I': 113.08406,
        'L': 113.08406,
        'K': 128.09496,
        'M': 131.04049,
        'F': 147.06841,
        'P': 97.05276,
        'S': 87.03203,
        'T': 101.04768,
        'W': 186.07931,
        'Y': 163.06333,
        'V': 99.06841,
        'U': 150.953636,
        'O': 237.147727,
        }



import sys
from app.drivers.base import PepProttableDriver
from app.drivers.options import peptable_options

from app.readers import tsv as tsvreader

from app.dataformats import mzidtsv as psmh
from app.dataformats import peptable as peph 
from app.dataformats import prottable as proth

from app.actions.psmtable import isosummarize
from app.actions import psmtopeptable


class CreatePeptableDriver(PepProttableDriver):
    """Creates unique peptide table from MzidTSV table. Will not change
    q-values/PEP after filtration, so these should have been calculated
    for the peptides beforehand or rescored afterwards.
    """
    outsuffix = '_peptable.tsv'
    command = 'peptides'
    commandhelp = ('Create peptide table from PSM TSV input, uses best '
                   'scoring PSM for each peptide and strips PSM '
                   'isobaric quant information. Retains MS1 quant info if '
                   'specified, by taking the highest precursor quant value '
                   'for a peptide. Optionally add isobaric quant and  modeled '
                   'q-values')

    def __init__(self):
        super().__init__()
        self.infiletype = 'TSV PSM table (MSGF+)'

    def set_options(self):
        super().set_options()
        self.options.update(self.define_options(['spectracol', 'scorecolpattern', 
            'quantcolpattern', 'precursorquantcolpattern', 'minint', 'denomcols',
            'denompatterns', 'mediansweep', 'medianintensity', 'median_or_avg',
            'logisoquant', 'mediannormalize', 'mednorm_factors', 'modelqvals',
            'qvalthreshold', 'keep_psms_na', 'minpeptidenr', 'totalprotfn'], peptable_options))

    def prepare(self):
        self.oldheader = tsvreader.get_tsv_header(self.fn)
        self.get_column_header_for_number(['spectracol'])
        self.scorecol = tsvreader.get_cols_in_file(self.scorecolpattern,
                                                   self.oldheader, True)
        self.precurquantcol = psmtopeptable.get_quantcols(self.precursorquantcolpattern,
                                                 self.oldheader, 'precur')


    def set_features(self):
        header = [x for x in self.oldheader if x != psmh.HEADER_SPECFILE]
        qpat = '[a-z]+[0-9]+plex_'
        try:
            isocols = tsvreader.get_columns_by_pattern(header, qpat)
        except RuntimeError:
            isocols = []
            if self.quantcolpattern:
                isocols = tsvreader.get_columns_by_pattern(header, self.quantcolpattern)
        for col in isocols:
            header.pop(header.index(col))
        if self.precurquantcol:
            header = [peph.HEADER_AREA if x == self.precurquantcol
                      else x for x in header]
        header = [peph.HEADER_PEPTIDE, peph.HEADER_LINKED_PSMS] + [
                x for x in header if x != psmh.HEADER_PEPTIDE]
        switch_map = {old: new for old, new in zip(
            [psmh.HEADER_PEPTIDE, psmh.HEADER_PROTEIN, psmh.HEADER_PEPTIDE_Q],
            [peph.HEADER_PEPTIDE, peph.HEADER_PROTEINS, peph.HEADER_QVAL])}
        self.header = [switch_map[field] if field in switch_map else field
                for field in header]
        peptides = psmtopeptable.generate_peptides(self.fn, self.oldheader,
                switch_map, self.scorecol, self.precurquantcol, self.spectracol)
        # Remove quant data if not specified any way to summarize
        if self.quantcolpattern and any([self.denomcols, self.denompatterns,
                self.mediansweep, self.medianintensity]):
            denomcols = False
            if self.denomcols is not None:
                denomcols = [self.number_to_headerfield(col, self.oldheader)
                             for col in self.denomcols]
            elif self.denompatterns is not None:
                denomcolnrs = [tsvreader.get_columns_by_combined_patterns(self.oldheader,
                    [self.quantcolpattern, pattern]) for pattern in self.denompatterns]
                denomcols = set([col for cols in denomcolnrs for col in cols])
            quantcols = tsvreader.get_columns_by_pattern(self.oldheader,
                                                   self.quantcolpattern)
            totalproteome, tpacc, tp_pepacc = False, False, False
            if self.totalprotfn:
                pep_tp_accs = [psmh.HEADER_MASTER_PROT, psmh.HEADER_SYMBOL,
                        psmh.HEADER_GENE, peph.HEADER_PROTEINS]
                totalphead = tsvreader.get_tsv_header(self.totalprotfn)
                totalpfield_found = False
                for tpacc, tp_pepacc in zip(proth.TPROT_HEADER_ACCS, pep_tp_accs):
                    if totalphead[0] == tpacc and tp_pepacc in self.header:
                        totalpfield_found = True
                        break
                if not totalpfield_found:
                    print('Could not find correct header field name in the total '
                            'proteome table passed. '
                            'Should be one of {}'.format(proth.TPROT_HEADER_ACCS))
                    sys.exit(1)
                totalproteome = tsvreader.generate_split_tsv_lines(self.totalprotfn, totalphead)
            mn_factors = False
            if self.mednorm_factors:
                mnhead = tsvreader.get_tsv_header(self.mednorm_factors)
                mn_factors = tsvreader.generate_split_tsv_lines(self.mednorm_factors, mnhead)
            nopsms = [isosummarize.get_no_psms_field(qf) for qf in quantcols]
            self.header = self.header + quantcols + nopsms + [proth.HEADER_NO_FULLQ_PSMS]
            peptides = isosummarize.get_isobaric_ratios(self.fn, self.oldheader, 
                    quantcols, denomcols, self.mediansweep, self.medianintensity,
                    self.median_or_avg, self.minint, peptides, self.header[0],
                    psmh.HEADER_PEPTIDE, totalproteome, tpacc, tp_pepacc,
                    self.logisoquant, self.mediannormalize, mn_factors, self.keepnapsms)
        if self.modelqvals:
            qix = self.header.index(peph.HEADER_QVAL) + 1
            self.header = self.header[:qix] + [peph.HEADER_QVAL_MODELED] + self.header[qix:]
            scorecol = tsvreader.get_cols_in_file(self.scorecolpattern,
                    self.oldheader, True)
            peptides = psmtopeptable.recalculate_qvals_linear_model(peptides, 
                    scorecol, self.qvalthreshold, self.minpeptidenr)
        self.features = peptides



import os

from app.readers import tsv as tsvreader
from app.dataformats import prottable as prottabledata


def create_lookup(fns, pqdb, poolnames, featcolnr, ms1_qcolpattern,
        isobqcolpattern, fdrcolpattern, pepcolpattern, flrcolpattern):
    psmnrpattern = prottabledata.HEADER_NO_PSMS_SUFFIX
    fullpsmpattern = prottabledata.HEADER_NO_FULLQ_PSMS
    patterns = [ms1_qcolpattern, fdrcolpattern, pepcolpattern, fullpsmpattern,
            flrcolpattern]
    storefuns = [pqdb.store_precursor_quants, pqdb.store_fdr, pqdb.store_pep,
            pqdb.store_fullq_psms, pqdb.store_ptm_flr]
    tablefn_map = create_tablefn_map(fns, pqdb, poolnames)
    feat_map = pqdb.get_feature_map()
    for pattern, storefun in zip(patterns, storefuns):
        if pattern is None:
            continue
        colmap = get_colmap(fns, pattern, single_col=True)
        if colmap:
            store_single_col_data(fns, tablefn_map, feat_map, storefun, colmap)
    if isobqcolpattern:
        isocolmap = get_colmap(fns, isobqcolpattern, antipattern=psmnrpattern)
        psmcolmap = get_colmap(fns, psmnrpattern)
        create_isobaric_quant_lookup(fns, tablefn_map, feat_map, pqdb, featcolnr,
                isocolmap, psmcolmap)


def create_tablefn_map(fns, pqdb, poolnames):
    """Stores protein/peptide table names in DB, returns a map with their
    respective DB IDs"""
    poolmap = {name: pid for (name, pid) in pqdb.get_all_poolnames()}
    pqdb.store_table_files([(poolmap[pool], os.path.basename(fn))
                            for fn, pool in zip(fns, poolnames)])
    return pqdb.get_tablefn_map()


def get_colmap(fns, pattern, single_col=False, antipattern=False):
    """For table files, loops through headers and checks which column(s)
    match a passed pattern. Those column(s) names are returned in a map with
    filenames as keys"""
    colmap = {}
    for fn in fns:
        header = tsvreader.get_tsv_header(fn)
        basefn = os.path.basename(fn)
        try:
            cols = tsvreader.get_cols_in_file(pattern, header, single_col)
        except RuntimeError:
            # Columns are not in this file
            cols = []
        if antipattern:
            try:
                anticols = tsvreader.get_cols_in_file(antipattern, header,
                                                      single_col)
            except RuntimeError:
                # The filtering "anti"-columns are not in the file, 
                anticols = []
            cols = [col for col in cols if col not in anticols]
        if cols:
            colmap[basefn] = cols
        else:
            return False
    return colmap


def store_single_col_data(fns, prottable_id_map, pacc_map, pqdbmethod, colmap):
    """General method to store single column data from protein tables
    in lookup"""
    to_store = []
    for fn, header, pquant in tsvreader.generate_tsv_pep_protein_quants(fns):
        pacc_id = pacc_map[pquant[header[0]]]
        pqdata = (pacc_id, prottable_id_map[fn], pquant[colmap[fn]])
        to_store.append(pqdata)
        if len(to_store) > 10000:
            pqdbmethod(to_store)
            to_store = []
    pqdbmethod(to_store)


def create_isobaric_quant_lookup(fns, tablefn_map, featmap, pqdb,
                                 featcolnr, allquantcols, psmcolmap):
    """Creates a lookup dict from peptide/protein quant input files and some
    input parameters. This assumes the order of quant columns and
    number-of-PSM columns is the same."""
    pqdb.store_quant_channels(
            map_psmnrcol_to_quantcol(allquantcols, psmcolmap, tablefn_map),
            psmcolmap)
    quantmap = pqdb.get_quantchannel_map()
    to_store = []
    for fn, header, pquant in tsvreader.generate_tsv_pep_protein_quants(fns):
        pqdata = get_isob_quant_data(pquant, header[featcolnr], tablefn_map[fn],
                                     featmap, quantmap)
        to_store.extend(pqdata)
        if len(to_store) > 10000:
            pqdb.store_isobaric_quants(to_store, psmcolmap)
            to_store = []
    pqdb.store_isobaric_quants(to_store, psmcolmap)


def map_psmnrcol_to_quantcol(quantcols, psmcols, tablefn_map):
    """This function yields tuples of table filename, isobaric quant column
    and if necessary number-of-PSM column"""
    if not psmcols:
        for fn in quantcols:
            for qcol in quantcols[fn]:
                yield (tablefn_map[fn], qcol)
    else:
        for fn in quantcols:
            for qcol, psmcol in zip(quantcols[fn], psmcols[fn]):
                yield (tablefn_map[fn], qcol, psmcol)


def get_isob_quant_data(pquant, featfield, fnid, featmap, qmap):
    """Turns a dict from a line of protein/peptide quant data into a set of
    tuples that can be stored in db"""
    feat_dbid = featmap[pquant[featfield]]
    for channel_name, (channel_id, psmfield) in qmap[fnid].items():
        if psmfield is None:
            yield (feat_dbid, channel_id, pquant[channel_name])
        else:
            yield (feat_dbid, channel_id, pquant[channel_name],
                   pquant[psmfield])


def build_proteintable(pqdb, mergecutoff, isobaric=False):
    """Fetches proteins and quants from joined lookup table, loops through
    them and when all of a protein's quants have been collected, yields the
    protein quant information."""
    # First get a dict with the data which is the same over sets (protein-gene mappings etc)
    pdmap = pqdb.create_pdata_map()
    previousfeat, outfeat = False, {}
    # Then loop over the SQLite output
    fdrfieldnr = pqdb.singlefields.index(prottabledata.HEADER_QVAL)
    # multi set output, means that a feature from DB can be delivered in
    # multiple lines (1/set). As soon as a new feat is received from db,
    # yield collected output and flush it for population with new protein
    for setname, feat_id, *setfeatvals in pqdb.merge_features():
        if feat_id != previousfeat and previousfeat and outfeat != {}:
            outfeat.update(pdmap[previousfeat])
            yield outfeat
            outfeat = {}
        if protein_pool_fdr_cutoff_ok(setfeatvals[fdrfieldnr], mergecutoff, feat_id, setname):
            for ix, field in enumerate(pqdb.singlefields):
                fieldname = '{}_{}'.format(setname, field)
                outfeat[fieldname] = setfeatvals[ix]
            if isobaric:
                channels = setfeatvals[len(pqdb.singlefields)].split(',')
                quants = setfeatvals[len(pqdb.singlefields)+1].split(',')
                try: 
                    ampsms = setfeatvals[len(pqdb.singlefields)+2].split(',')
                except AttributeError:
                    ampsms = [0] * len(channels)
                for ch, quant, ampsm in zip(channels, quants, ampsms):
                    outfeat['{}_{}'.format(setname, ch)] = quant
                    if ampsms:
                        outfeat['{}_{}{}'.format(setname, ch, prottabledata.HEADER_NO_PSMS_SUFFIX)] = ampsm
        previousfeat = feat_id
    if outfeat != {}:
        outfeat.update(pdmap[feat_id])
        yield outfeat



def protein_pool_fdr_cutoff_ok(fdr, fdrcutoff, feature, setname):
    if not fdrcutoff:
        return True
    warning = False
    if fdr is False:
        warning = True
    try:
        fdr = float(fdr)
    except (TypeError, ValueError):
        warning = True
    if warning:
        print('WARNING, filtering merge table on FDR but feat ID {} in '
                'set {} has FDR value "{}" in '
                'lookup.'.format(feature, setname, fdr))
        return False
    return fdr < fdrcutoff



from app.lookups.sqlite.protpeptable import ProtPepTable
from app.dataformats import prottable as ph


class ProtGeneTableBase(ProtPepTable):
    stdheaderfields = [
            ph.HEADER_NO_PSM,
            ph.HEADER_NO_PEPTIDE,
            ph.HEADER_NO_UNIPEP,
            ]
    singlefields = stdheaderfields + [ph.HEADER_QVAL, ph.HEADER_AREA, ph.HEADER_NO_FULLQ_PSMS]

    def add_tables(self, tabletypes=[]):
        ttypes = ['fntable', 'isoqtable', 'isochtable', 'prectable',
                'fdrtable', 'fullqpsmtable', 'flrtable']
        self.create_tables([self.table_map[self.datatype][x] for x in ttypes
            if x in self.table_map[self.datatype]])

    

class ProtTableDB(ProtGeneTableBase):
    datatype = 'protein'
    colmap = {'protein_group_master': ['master_id', 'protein_acc'],
              'proteins': ['pacc_id', 'protein_acc'],
              'protein_precur_quanted': ['pacc_id', 'prottable_id', 'quant'],
              'protein_fdr': ['pacc_id', 'prottable_id', 'fdr'],
              'protein_iso_fullpsms': ['pacc_id', 'prottable_id', 'amount_psms'],
              'protquant_channels': ['channel_id', 'prottable_id',
                                     'channel_name', 'amount_psms_name'],
              'protein_iso_quanted': ['proteinquant_id', 'pacc_id',
                                      'channel_id', 'quantvalue',
                                      'amount_psms'],
              }

    def create_pdata_map(self):
        """This runs only once, returns the data which is not dependent on sets,
        in a dict with accessions as keys"""
        sql = """
SELECT pgm.master_id, p.protein_acc, IFNULL(g.gene_acc, 'NA'),
    IFNULL(aid.assoc_id, 'NA'), cov.coverage, sub.pgc, 
    sub.pgcnr, IFNULL(pd.description, 'NA') FROM protein_group_master AS pgm
    LEFT OUTER JOIN (
            SELECT master_id, GROUP_CONCAT(protein_acc) AS pgc, 
            COUNT(protein_acc) AS pgcnr FROM protein_group_content 
            GROUP BY master_id
            ) AS sub ON sub.master_id=pgm.master_id 
    INNER JOIN proteins AS p ON pgm.pacc_id=p.pacc_id
    LEFT OUTER JOIN protein_coverage AS cov ON 
    p.protein_acc=cov.protein_acc 
    LEFT OUTER JOIN ensg_proteins AS egp ON pgm.pacc_id=egp.pacc_id
    LEFT OUTER JOIN genes AS g ON egp.gene_id=g.gene_id
    LEFT OUTER JOIN genename_proteins AS gnp ON pgm.pacc_id=gnp.pacc_id
    LEFT OUTER JOIN associated_ids AS aid ON aid.gn_id=gnp.gn_id
    LEFT OUTER JOIN prot_desc AS pd ON pd.pacc_id=p.pacc_id
    """
        cursor = self.get_cursor()
        pgdata = {}
        for mid, macc, gacc, aid, cov, pgc, pgnr, desc in cursor.execute(sql):
            pgdata[mid] = {
                    ph.HEADER_PROTEIN: macc,
                    ph.HEADER_GENEID: gacc,
                    ph.HEADER_GENENAME: aid,
                    ph.HEADER_COVERAGE: cov,
                    ph.HEADER_CONTENTPROT: pgc,
                    ph.HEADER_NO_PROTEIN: pgnr,
                    ph.HEADER_DESCRIPTION: desc,
                    }
        return pgdata

    def merge_features(self):
        sql = """
    SELECT bs.set_name, pgm.master_id, COUNT(DISTINCT ppg.psm_id), 
    COUNT (DISTINCT ps.pep_id), COUNT(DISTINCT uni.pep_id), 
    IFNULL(pf.fdr, 'NA'), ppq.quant, fqpsm.amount_psms, GROUP_CONCAT(pqc.channel_name), GROUP_CONCAT(piq.quantvalue),
    GROUP_CONCAT(piq.amount_psms)
        FROM psm_protein_groups AS ppg 
        INNER JOIN psms ON ppg.psm_id=psms.psm_id 
        INNER JOIN mzml ON psms.spectra_id=mzml.spectra_id
        INNER JOIN mzmlfiles ON mzml.mzmlfile_id=mzmlfiles.mzmlfile_id
        INNER JOIN biosets AS bs ON mzmlfiles.set_id=bs.set_id
        INNER JOIN peptide_sequences AS ps ON psms.pep_id=ps.pep_id
        INNER JOIN protein_tables AS pt ON pt.set_id=bs.set_id
        INNER JOIN protein_group_master AS pgm ON pgm.master_id=ppg.master_id
        INNER JOIN proteins AS prots ON prots.pacc_id=pgm.pacc_id
        LEFT OUTER JOIN protein_fdr AS pf ON pf.prottable_id=pt.prottable_id AND 
            pf.pacc_id=prots.pacc_id
        LEFT OUTER JOIN protein_precur_quanted AS ppq ON ppq.prottable_id=pt.prottable_id AND 
            ppq.pacc_id=prots.pacc_id
        LEFT OUTER JOIN protein_iso_fullpsms AS fqpsm ON fqpsm.prottable_id=pt.prottable_id AND 
            fqpsm.pacc_id=prots.pacc_id
        LEFT OUTER JOIN protquant_channels AS pqc ON pqc.prottable_id=pt.prottable_id
        LEFT OUTER JOIN protein_iso_quanted AS piq ON piq.channel_id=pqc.channel_id AND
            piq.pacc_id=prots.pacc_id
        LEFT OUTER JOIN (
                SELECT ppg.pep_id AS pep_id FROM (
                    SELECT psms.pep_id AS pep_id, COUNT (DISTINCT ppg.master_id) AS nrpep 
                        FROM psm_protein_groups AS ppg INNER JOIN psms USING(psm_id)
                        GROUP BY psms.pep_id
                    ) AS ppg WHERE ppg.nrpep==1
                ) AS uni ON uni.pep_id=ps.pep_id
        GROUP BY pgm.master_id, bs.set_id
        """
        cursor = self.get_cursor()
        cursor.execute(sql)
        return cursor


class ProtTableNoGroupNoMapDB(ProtGeneTableBase):
    datatype = 'protein'
    colmap = {#'protein_group_master': ['master_id', 'protein_acc'],
              'proteins': ['pacc_id', 'protein_acc'],
              'protein_precur_quanted': ['pacc_id', 'prottable_id', 'quant'],
              'protein_fdr': ['pacc_id', 'prottable_id', 'fdr'],
              'protein_iso_fullpsms': ['pacc_id', 'prottable_id', 'amount_psms'],
              'protquant_channels': ['channel_id', 'prottable_id',
                                     'channel_name', 'amount_psms_name'],
              'protein_iso_quanted': ['proteinquant_id', 'pacc_id',
                                      'channel_id', 'quantvalue',
                                      'amount_psms'],
              }

    def create_pdata_map(self):
        """This runs only once, returns the data which is not dependent on sets,
        in a dict with accessions as keys"""
        sql = 'SELECT pacc_id, protein_acc FROM proteins'
        cursor = self.get_cursor()
        return {pid: {ph.HEADER_PROTEIN: pacc} for pid, pacc in cursor.execute(sql)}

    def merge_features(self):
        sql = """
    SELECT bs.set_name, prots.pacc_id, COUNT(DISTINCT pp.psm_id), 
    COUNT (DISTINCT ps.pep_id), COUNT(DISTINCT uni.pep_id), 
    IFNULL(pf.fdr, 'NA'), ppq.quant, fqpsm.amount_psms, GROUP_CONCAT(pqc.channel_name), GROUP_CONCAT(piq.quantvalue),
    GROUP_CONCAT(piq.amount_psms)
        FROM protein_psm AS pp 
        INNER JOIN psms ON pp.psm_id=psms.psm_id 
        INNER JOIN mzml ON psms.spectra_id=mzml.spectra_id
        INNER JOIN mzmlfiles ON mzml.mzmlfile_id=mzmlfiles.mzmlfile_id
        INNER JOIN biosets AS bs ON mzmlfiles.set_id=bs.set_id
        INNER JOIN peptide_sequences AS ps ON psms.pep_id=ps.pep_id
        INNER JOIN protein_tables AS pt ON pt.set_id=bs.set_id
        INNER JOIN proteins AS prots ON prots.protein_acc=pp.protein_acc
        LEFT OUTER JOIN protein_fdr AS pf ON pf.prottable_id=pt.prottable_id AND 
            pf.pacc_id=prots.pacc_id
        LEFT OUTER JOIN protein_precur_quanted AS ppq ON ppq.prottable_id=pt.prottable_id AND 
            ppq.pacc_id=prots.pacc_id
        LEFT OUTER JOIN protein_iso_fullpsms AS fqpsm ON fqpsm.prottable_id=pt.prottable_id AND 
            fqpsm.pacc_id=prots.pacc_id
        LEFT OUTER JOIN protquant_channels AS pqc ON pqc.prottable_id=pt.prottable_id
        LEFT OUTER JOIN protein_iso_quanted AS piq ON piq.channel_id=pqc.channel_id AND
            piq.pacc_id=prots.pacc_id
        LEFT OUTER JOIN (
                SELECT ppg.pep_id AS pep_id FROM (
                    SELECT psms.pep_id AS pep_id, COUNT (DISTINCT pp.protein_acc) AS nrpep 
                        FROM protein_psm AS pp INNER JOIN psms USING(psm_id)
                        GROUP BY psms.pep_id
                    ) AS ppg WHERE ppg.nrpep==1
                ) AS uni ON uni.pep_id=ps.pep_id
        GROUP BY prots.pacc_id, bs.set_id
        """
        cursor = self.get_cursor()
        cursor.execute(sql)
        return cursor



class GeneTableDB(ProtGeneTableBase):
    datatype = 'gene'
    colmap = {'genes': ['gene_id', 'gene_acc'],
              'gene_precur_quanted': ['gene_id', 'genetable_id', 'quant'],
              'gene_fdr': ['gene_id', 'genetable_id', 'fdr'],
              'gene_iso_fullpsms': ['gene_id', 'genetable_id', 'amount_psms'],
              'genequant_channels': ['channel_id', 'genetable_id',
                                     'channel_name', 'amount_psms_name'],
              'gene_iso_quanted': ['genequant_id', 'gene_id',
                                   'channel_id', 'quantvalue', 'amount_psms'],
              }

    def create_pdata_map(self):
        """This runs only once, returns the data which is not dependent on sets,
        in a dict with accessions as keys"""
        sql = """
SELECT g.gene_acc, GROUP_CONCAT(p.protein_acc, ';'), IFNULL(aid.assoc_id, 'NA'), 
    IFNULL(pd.description, 'NA')
    FROM genes AS g
    LEFT OUTER JOIN ensg_proteins AS egp ON egp.gene_id=g.gene_id
    LEFT OUTER JOIN proteins AS p ON p.pacc_id=egp.pacc_id
    LEFT OUTER JOIN genename_proteins AS gnp ON gnp.pacc_id=egp.pacc_id
    LEFT OUTER JOIN associated_ids AS aid ON aid.gn_id=gnp.gn_id
    LEFT OUTER JOIN prot_desc AS pd ON pd.pacc_id=p.pacc_id
    GROUP BY g.gene_acc
    """
        cursor = self.get_cursor()
        pgdata = {}
        for gacc, pacc, aid, desc in cursor.execute(sql):
            pgdata[gacc] = {
                    ph.HEADER_PROTEINS: pacc,
                    ph.HEADER_GENEID: gacc,
                    ph.HEADER_GENENAME: aid,
                    ph.HEADER_DESCRIPTION: desc,
                    }
        return pgdata

    def merge_features(self):
  ### protein_acc on gene table is indexed??
        sql = """
    SELECT bs.set_name, g.gene_acc, COUNT(DISTINCT ppsm.psm_id), 
    COUNT (DISTINCT ps.pep_id), COUNT(DISTINCT uniq.pep_id), 
    IFNULL(gf.fdr, 'NA'), gpq.quant, fqpsm.amount_psms,
    GROUP_CONCAT(gqc.channel_name), GROUP_CONCAT(giq.quantvalue),
    GROUP_CONCAT(giq.amount_psms)
        FROM protein_psm AS ppsm
        INNER JOIN psms ON ppsm.psm_id=psms.psm_id 
        INNER JOIN mzml ON psms.spectra_id=mzml.spectra_id
        INNER JOIN mzmlfiles ON mzml.mzmlfile_id=mzmlfiles.mzmlfile_id
        INNER JOIN biosets AS bs ON mzmlfiles.set_id=bs.set_id
        INNER JOIN peptide_sequences AS ps ON psms.pep_id=ps.pep_id
        INNER JOIN gene_tables AS gt ON gt.set_id=bs.set_id
        INNER JOIN proteins AS p ON p.protein_acc=ppsm.protein_acc
        INNER JOIN ensg_proteins AS egp ON egp.pacc_id=p.pacc_id
        INNER JOIN genes AS g ON g.gene_id=egp.gene_id
        INNER JOIN gene_fdr AS gf ON gf.genetable_id=gt.genetable_id AND 
            gf.gene_id=g.gene_id
        LEFT OUTER JOIN gene_precur_quanted AS gpq ON gpq.genetable_id=gt.genetable_id AND 
            gpq.gene_id=g.gene_id
        LEFT OUTER JOIN gene_iso_fullpsms AS fqpsm ON fqpsm.genetable_id=gt.genetable_id AND 
            fqpsm.gene_id=g.gene_id
        LEFT OUTER JOIN genequant_channels AS gqc ON gqc.genetable_id=gt.genetable_id
        LEFT OUTER JOIN gene_iso_quanted AS giq ON giq.channel_id=gqc.channel_id AND
            giq.gene_id=g.gene_id

        LEFT OUTER JOIN (
                SELECT psmg.pep_id AS pep_id FROM (
                    SELECT psms.pep_id AS pep_id, COUNT (DISTINCT g.gene_acc) AS nrpep 
                        FROM protein_psm AS ppsm INNER JOIN psms USING(psm_id)
                        INNER JOIN proteins AS p ON p.protein_acc=ppsm.protein_acc
                        INNER JOIN ensg_proteins AS egp ON egp.pacc_id=p.pacc_id
                        INNER JOIN genes AS g ON g.gene_id=egp.gene_id
                        GROUP BY psms.pep_id
                    ) AS psmg WHERE psmg.nrpep==1
                ) AS uniq ON uniq.pep_id=ps.pep_id
        GROUP BY g.gene_id, bs.set_id
        """
        cursor = self.get_cursor()
        cursor.execute(sql)
        return cursor


class GeneTableAssocIDsDB(GeneTableDB):
    datatype = 'assoc'
    colmap = {'associated_ids': ['gn_id', 'assoc_id'],
              'assoc_precur_quanted': ['gn_id', 'genetable_id', 'quant'],
              'assoc_fdr': ['gn_id', 'genetable_id', 'fdr'],
              'assoc_iso_fullpsms': ['gn_id', 'genetable_id', 'amount_psms'],
              'genequant_channels': ['channel_id', 'genetable_id',
                                     'channel_name', 'amount_psms_name'],
              'assoc_iso_quanted': ['genequant_id', 'gn_id',
                                   'channel_id', 'quantvalue', 'amount_psms'],
              }

    def create_pdata_map(self):
        """This runs only once, returns the data which is not dependent on sets,
        in a dict with accessions as keys"""
        sql = """
SELECT gn.assoc_id, GROUP_CONCAT(p.protein_acc, ';'), IFNULL(g.gene_acc, 'NA'), 
    IFNULL(pd.description, 'NA')
    FROM associated_ids AS gn
    LEFT OUTER JOIN genename_proteins AS gnp ON gnp.gn_id=gn.gn_id
    LEFT OUTER JOIN proteins AS p ON p.pacc_id=gnp.pacc_id
    LEFT OUTER JOIN ensg_proteins AS egp ON egp.pacc_id=gnp.pacc_id
    LEFT OUTER JOIN genes AS g ON g.gene_id=egp.gene_id
    LEFT OUTER JOIN prot_desc AS pd ON pd.pacc_id=p.pacc_id
    GROUP BY gn.gn_id
    """
        cursor = self.get_cursor()
        pgdata = {}
        for aid, pacc, gacc, desc in cursor.execute(sql):
            pgdata[aid] = {
                    ph.HEADER_PROTEINS: pacc,
                    ph.HEADER_GENEID: gacc,
                    ph.HEADER_GENENAME: aid,
                    ph.HEADER_DESCRIPTION: desc,
                    }
        return pgdata

    def merge_features(self):
  ### protein_acc on gene table is indexed??
  ### check distinct stuff, pgroups etc need it, matched with content numbers (which cnanot habe it)
        sql = """
    SELECT bs.set_name, gn.assoc_id, COUNT(DISTINCT ppsm.psm_id), 
    COUNT (DISTINCT ps.pep_id), COUNT(DISTINCT uniq.pep_id), 
    IFNULL(gf.fdr, 'NA'), gpq.quant, fqpsm.amount_psms,
    GROUP_CONCAT(gqc.channel_name), GROUP_CONCAT(giq.quantvalue),
    GROUP_CONCAT(giq.amount_psms)
        FROM protein_psm AS ppsm
        INNER JOIN psms ON ppsm.psm_id=psms.psm_id 
        INNER JOIN mzml ON psms.spectra_id=mzml.spectra_id
        INNER JOIN mzmlfiles ON mzml.mzmlfile_id=mzmlfiles.mzmlfile_id
        INNER JOIN biosets AS bs ON mzmlfiles.set_id=bs.set_id
        INNER JOIN peptide_sequences AS ps ON psms.pep_id=ps.pep_id
        INNER JOIN gene_tables AS gt ON gt.set_id=bs.set_id
        INNER JOIN proteins AS p ON p.protein_acc=ppsm.protein_acc
        INNER JOIN genename_proteins AS gnp ON gnp.pacc_id=p.pacc_id
        INNER JOIN associated_ids AS gn ON gn.gn_id=gnp.gn_id
        INNER JOIN assoc_fdr AS gf ON gf.genetable_id=gt.genetable_id AND 
            gf.gn_id=gn.gn_id
        LEFT OUTER JOIN assoc_precur_quanted AS gpq ON gpq.genetable_id=gt.genetable_id AND 
            gpq.gn_id=gn.gn_id
        LEFT OUTER JOIN assoc_iso_fullpsms AS fqpsm ON fqpsm.genetable_id=gt.genetable_id AND 
            fqpsm.gn_id=gn.gn_id
        LEFT OUTER JOIN genequant_channels AS gqc ON gqc.genetable_id=gt.genetable_id
        LEFT OUTER JOIN assoc_iso_quanted AS giq ON giq.channel_id=gqc.channel_id AND
            giq.gn_id=gn.gn_id

        LEFT OUTER JOIN (
                SELECT psmg.pep_id AS pep_id FROM (
                    SELECT psms.pep_id AS pep_id, COUNT (DISTINCT gn.assoc_id) AS nrpep 
                        FROM protein_psm AS ppsm INNER JOIN psms USING(psm_id)
                        INNER JOIN proteins AS p ON p.protein_acc=ppsm.protein_acc
                        INNER JOIN genename_proteins AS gnp ON gnp.pacc_id=p.pacc_id
                        INNER JOIN associated_ids AS gn ON gn.gn_id=gnp.gn_id
                        GROUP BY psms.pep_id
                    ) AS psmg WHERE psmg.nrpep==1
                ) AS uniq ON uniq.pep_id=ps.pep_id
        GROUP BY gn.gn_id, bs.set_id
        """
        cursor = self.get_cursor()
        cursor.execute(sql)
        return cursor



from lxml import etree
import re

from app.readers import xmlformatting as formatting
from app.readers import percolator as reader


def protein_header_split_generator(elements, ns, can_headers, headers):
    """Loop through proteins of each PSM/peptide. If a protein does not
    match any of headers, discard PSM/peptide immediately"""
    for el in elements:
        header_matching = False
        can = False
        for protein in el.findall('{%s}protein_id' % ns['xmlns']):
            if any(re.search(h, protein.text) for h in can_headers):
                can = True
                break #as soon as a canonical match was found break
            """for classes other than known,
               check if there is at least one protein matching the specified header
               and those with matches to the canonical proteins will not be used"""
            if any(re.search(h, protein.text) for h in headers):
                    header_matching = True
        if (header_matching and not can) or ((headers == can_headers) and can):
            yield formatting.string_and_clear(el, ns)
        else:
            formatting.clear_el(el)


def split_protein_header_id_type(elements, ns, protheaders):
    can_headers = protheaders.split('|')[0].split(':')[-1].strip(';').split(';')
    other_headers = protheaders.split('|')[-1].split(':')[-1].strip(';').split(';')
    return {x: protein_header_split_generator(elements[x], ns, can_headers, other_headers)
            for x in ['psm', 'peptide']}




import os
from decimal import Decimal
from collections import defaultdict

from app.readers import openms as openmsreader
DB_STORE_CHUNK = 500000
FEATURE_ALIGN_WINDOW_AMOUNT = 1000
PROTON_MASS = 1.0072


def create_isobaric_quant_lookup(quantdb, specfns, isobaricfns):
    """Creates an sqlite lookup table of scannrs with quant data.

    spectra - an iterable of tupled (filename, spectra)
    consensus_els - a iterable with consensusElements"""
    mzmlmap = quantdb.get_mzmlfile_map()
    for specfpath, isofn in zip(specfns, isobaricfns):
        specfn = os.path.basename(specfpath)
        quants, pifs = [], []
        # store quantchannels in lookup and generate a db_id vs channel map
        channelmap = openmsreader.get_quantmap(isofn)
        channels_store = ((name,) for name, c_id
                          in sorted(channelmap.items(), key=lambda x: int(x[1])))
        quantdb.store_channelmap(channels_store)
        print(channelmap)
        channelmap_dbid = {channelmap[ch_name]: ch_id for ch_id, ch_name in
                quantdb.get_channelmap() if ch_name in channelmap}
        # Get DB spectra IDs keyed by scan SIDs
        specmap = quantdb.get_specmap(mzmlmap[specfn])
        for _, consensus_el in openmsreader.mzmlfn_cons_el_generator([specfn], [isofn]):
            pif, sid = openmsreader.get_consxml_sid_pif(consensus_el)
            # FIXME what if sid not in specmap? error then
            spectra_id = specmap[sid]['id']
            if pif:
                pifs.append((spectra_id, float(pif)))
            qdata = get_quant_data(consensus_el)
            for channel_no in sorted(qdata.keys()):
                quants.append((spectra_id, channelmap_dbid[channel_no],
                               qdata[channel_no]))
                if len(quants) == DB_STORE_CHUNK:
                    quantdb.store_isobaric_quants(quants, pifs)
                    quants, pifs = [], []
        quantdb.store_isobaric_quants(quants, pifs)
    quantdb.index_isobaric_quants()


def create_precursor_quant_lookup(quantdb, mzmlfn_feats, sum_or_apex, quanttype,
        rttol, mztol, mztoltype):
    """Fills quant sqlite with precursor quant from:
        features - generator of xml features from openms
    """
    featparsermap = {'kronik': kronik_featparser,
                     'openms': openms_featparser,
                     'dinosaur': dinosaur_featparser,
                     }
    features, fwhms = [], []
    mzmlmap = quantdb.get_mzmlfile_map()
    for specfn, feat_element in mzmlfn_feats:
        feat = featparsermap[quanttype](feat_element, sum_or_apex)
        features.append((mzmlmap[specfn], feat['rt'], feat['mz'],
                         feat['charge'], feat['intensity'])
                        )
        fwhms.append(feat['fwhm'])
        if len(features) == DB_STORE_CHUNK:
            id_feats = quantdb.store_ms1_quants(features)
            if quanttype == 'dinosaur':
                quantdb.store_fwhm(zip([x[0] for x in id_feats], fwhms))
            features, fwhms = [], []
    id_feats = quantdb.store_ms1_quants(features)
    if quanttype == 'dinosaur':
        quantdb.store_fwhm(zip([x[0] for x in id_feats], fwhms))
    quantdb.index_precursor_quants()
    align_quants_psms(quantdb, rttol, mztol, mztoltype)
    quantdb.index_aligned_quants()


def get_minmax(center, tolerance, toltype=None):
    center = float(center)
    if toltype == 'ppm':
        tolerance = int(tolerance) / 1000000 * center
    else:
        tolerance = float(tolerance)
    return center - tolerance, center + tolerance


def align_quants_psms(quantdb, rt_tolerance, mz_tolerance, mz_toltype):
    allspectra = quantdb.get_spectra_mz_sorted()
    featwindow_max_mz = -1
    spec_feat_store = []
    active_fn = None
    for spec_id, fn_id, charge, mz, rt in allspectra:
        if fn_id != active_fn:
            active_fn = fn_id
            featwindow_max_mz = -1
            fnfeats = quantdb.get_fnfeats(fn_id)
        minmz, maxmz = get_minmax(mz, mz_tolerance, mz_toltype)
        if maxmz > featwindow_max_mz:
            feat_map, featwindow_max_mz = get_precursors_from_window(fnfeats,
                                                                     minmz)
        best_feat_id = align_psm(mz, rt, charge, feat_map, rt_tolerance)
        if not best_feat_id:
            continue
        spec_feat_store.append((spec_id, best_feat_id))
        if len(spec_feat_store) > DB_STORE_CHUNK:
            quantdb.store_ms1_alignments(spec_feat_store)
            spec_feat_store = []
    quantdb.store_ms1_alignments(spec_feat_store)


def align_psm(psm_mz, psm_rt, charge, featmap, rttol):
    rttol_min = rttol / 60
    alignments = defaultdict(list)
    try:
        featlist = featmap[charge]
    except KeyError:
        return False
    for feat_mz, feat_rt, feat_id in featlist:
        rtdiff = abs(psm_rt - feat_rt)
        mzdiff = abs(psm_mz - feat_mz)
        if rtdiff > rttol_min:
            continue
        alignments[mzdiff].append((feat_id, rtdiff))
    try:
        best_mz = alignments[min(alignments)]
    except ValueError:
        return False
    if len(best_mz) > 1:
        print('Multiple features of same m/z found within RT window, possibly '
                'your RT tolerance is set too large')
        return sorted(best_mz, key=lambda x: x[1])[0][0]
    else:
        return best_mz[0][0]


def get_precursors_from_window(mzfeatmap, minmz):
    """Returns a dict of a specified amount of features from the
    ms1 quant database, and the highest mz of those features"""
    chargemap = {}
    mz, amount_feats = False, 0
    features = []
    for mz, feat_id, charge, rt in mzfeatmap:
        if mz < minmz:
            continue
        elif amount_feats > FEATURE_ALIGN_WINDOW_AMOUNT:
            break
        amount_feats += 1
        try:
            chargemap[charge].append((mz, rt, feat_id))
        except KeyError:
            chargemap[charge] = [(mz, rt, feat_id)]
    return chargemap, mz


def kronik_featparser(feature, sum_or_apex):
    charge = int(feature['Charge'])
    mz = (float(feature['Monoisotopic Mass']) + charge * PROTON_MASS) / charge
    intkey = {'sum': 'Summed Intensity', 'apex': 'Best Intensity'}[sum_or_apex]
    return {'rt': round(float(feature['Best RTime']), 12),
            'mz': mz,
            'charge': charge,
            'intensity': float(feature[intkey]),
            'fwhm': False,
            }


def dinosaur_featparser(feature, sum_or_apex):
    # FIXME is kronik, change, add FWHM
    #mz = (float(feature['mz']) + charge * PROTON_MASS) / charge
    intkey = {'sum': 'intensitySum', 'apex': 'intensityApex'}[sum_or_apex]
    return {'rt': round(float(feature['rtApex']), 12),
            'mz': float(feature['mz']),
            'charge': int(feature['charge']),
            'intensity': float(feature[intkey]),
            'fwhm': float(feature['fwhm']),
            }


def openms_featparser(feature):
    feat = openmsreader.get_feature_info(feature)
    feat['rt'] = round(float(Decimal(feat['rt']) / 60), 12)
    return feat


def get_quant_data(cons_el):
    """Gets quant data from consensusXML element"""
    quant_out = {}
    for reporter in cons_el.findall('.//element'):
        quant_out[reporter.attrib['map']] = reporter.attrib['it']
    return quant_out



from app.lookups.sqlite.base import ResultLookupInterface


# Indices that belong to positions of these features in output from
# function get_all_psms_proteingroups:
MASTER_INDEX = 1
PROTEIN_ACC_INDEX = 2
PEPTIDE_COUNT_INDEX = 3
PSM_COUNT_INDEX = 4
PROTEIN_SCORE_INDEX = 5
COVERAGE_INDEX = 6
EVIDENCE_LVL_INDEX = 7


class PSMDB(ResultLookupInterface):
    def add_tables(self, tabletypes):
        self.create_tables(['psms', 'psmrows', 'peptide_sequences', 'fastafn',
                            'proteins', 'protein_evidence', 'protein_seq',
                            'prot_desc', 'protein_psm', 'genes',
                            'associated_ids', 'ensg_proteins', 'genename_proteins',
                            'mods', 'modpos'])
        if 'proteingroup' in tabletypes:
            self.create_tables(['protein_coverage', 'protein_group_master',
                                'protein_group_content', 'psm_protein_groups'])

    def get_fasta_md5(self):
        cursor = self.get_cursor()
        cursor.execute('SELECT md5 FROM fastafn LIMIT 1')
        md5 = cursor.fetchone()
        return md5 if md5 is not None else False

    def store_fasta(self, fn, md5, prot, evids, seq, desc, ensg, symbols):
        cursor = self.get_cursor()
        cursor.execute('INSERT INTO fastafn(filename, md5) VALUES(?, ?)',
                (fn, md5))
        self.store_proteins(prot, evidence_lvls=evids, sequences=seq)
        protids = self.get_protids()
        cursor = self.get_cursor()
        cursor.executemany(
            'INSERT INTO prot_desc(pacc_id, description) '
            'VALUES(?, ?)', ((protids[x[0]], x[1]) for x in desc))
        self.conn.commit()
        self.index_column('protdesc_index', 'prot_desc', 'pacc_id')

        cursor = self.get_cursor()
        cursor.executemany(
            'INSERT INTO genes(gene_acc) VALUES(?)', set((x[0],) for x in ensg))
        self.conn.commit()
        ensgids = {x[0]: x[1] for x in cursor.execute(
            'SELECT g.gene_acc, g.gene_id FROM genes AS g')}
        ensg = [(ensgids[x[0]], protids[x[1]]) for x in ensg]
        cursor.executemany(
            'INSERT INTO ensg_proteins(gene_id, pacc_id) VALUES(?, ?)', ensg)
        self.conn.commit()
        self.index_column('ensg_p_ix', 'ensg_proteins', 'pacc_id')
        self.index_column('ensg_ensg_ix', 'ensg_proteins', 'gene_id')

        cursor = self.get_cursor()
        # symbols is list of sym/prot id, so unique the symbol ids and save
        cursor.executemany(
            'INSERT INTO associated_ids(assoc_id) VALUES(?)',
            set((x[0],) for x in symbols))
        symids = {x[0]: x[1] for x in cursor.execute(
            'SELECT s.assoc_id, s.gn_id FROM associated_ids AS s')}
        symbols = [(symids[x[0]], protids[x[1]]) for x in symbols]
        cursor.executemany(
            'INSERT INTO genename_proteins(gn_id, pacc_id) VALUES(?, ?)',
            symbols)
        self.conn.commit()
        self.index_column('gp_p_ix', 'genename_proteins', 'pacc_id')
        self.index_column('gp_gn_ix', 'genename_proteins', 'gn_id')

    def store_proteins(self, proteins, evidence_lvls=False, sequences=False):
        cursor = self.get_cursor()
        cursor.executemany(
            'INSERT INTO proteins(protein_acc) '
            'VALUES(?)', proteins)
        self.conn.commit()
        cursor = self.get_cursor()
        if evidence_lvls:
            cursor.executemany(
                'INSERT INTO protein_evidence(protein_acc, evidence_lvl) '
                'VALUES(?, ?)', evidence_lvls)
        if sequences:
            cursor.executemany(
                'INSERT INTO protein_seq(protein_acc, sequence) '
                'VALUES(?, ?)', sequences)
        self.conn.commit()
        self.index_column('proteins_index', 'proteins', 'protein_acc')
        self.index_column('evidence_index', 'protein_evidence', 'protein_acc')

    def get_protids(self):
        cur = self.get_cursor()
        return {x[0]: x[1] for x in cur.execute(
            'SELECT p.protein_acc, p.pacc_id FROM proteins AS p')}

    def get_protein_gene_map(self):
        cursor = self.get_cursor()
        cursor.execute(
            'SELECT p.protein_acc, g.gene_acc, aid.assoc_id, d.description '
            'FROM proteins AS p '
            'LEFT OUTER JOIN ensg_proteins AS ep ON p.pacc_id=ep.pacc_id '
            'LEFT OUTER JOIN genename_proteins AS gnp ON p.pacc_id=gnp.pacc_id '
            'LEFT OUTER JOIN genes AS g ON ep.gene_id=g.gene_id '
            'LEFT OUTER JOIN associated_ids AS aid ON gnp.gn_id=aid.gn_id '
            'LEFT OUTER JOIN prot_desc AS d ON p.pacc_id=d.pacc_id'
        )
        gpmap = {p_acc: {'gene': gene, 'symbol': sym, 'desc': desc}
                 for p_acc, gene, sym, desc in cursor}
        return gpmap

    def store_pepseqs(self, sequences):
        cursor = self.get_cursor()
        cursor.executemany(
            'INSERT INTO peptide_sequences(sequence) VALUES(?)', sequences)
        self.conn.commit()

    def store_psms(self, psms):
        cursor = self.get_cursor()
        cursor.executemany(
            'INSERT INTO psms(psm_id, pep_id, score, spectra_id) '
            'VALUES(?, ?, ?, ?)', ((psm['psm_id'], psm['seq'], psm['score'],
                                    psm['spec_id'])
                                   for psm in psms))
        cursor.executemany(
            'INSERT INTO psmrows(psm_id, rownr) VALUES(?, ?)',
            ((psm['psm_id'], psm['rownr']) for psm in psms))
        self.conn.commit()
    
    def get_highest_rownr(self):
        cursor = self.get_cursor()
        cursor.execute('SELECT MAX(rownr) FROM psmrows')
        return int(cursor.fetchone()[0])

    def store_peptides_proteins(self, allpepprot, psmids_to_store):
        ppmap = {psm_id: allpepprot[psm_id] for psm_id in psmids_to_store}
        prot_psm_ids = ((prot_acc, psm_id)
                        for psm_id, prots in ppmap.items()
                        for prot_acc in prots)
        cursor = self.get_cursor()
        cursor.executemany(
            'INSERT INTO protein_psm(protein_acc, psm_id)'
            ' VALUES (?, ?)', prot_psm_ids)
        self.conn.commit()

    def get_peptide_seq_map(self):
        cursor = self.get_cursor()
        seqs = cursor.execute('SELECT pep_id, sequence FROM peptide_sequences')
        return {seq: pepid for pepid, seq in seqs.fetchall()}

    def index_psms(self):
        self.index_column('psmid_index', 'psms', 'psm_id')
        self.index_column('psmspecid_index', 'psms', 'spectra_id')
        self.index_column('psmrowid_index', 'psmrows', 'psm_id')
        self.index_column('psmrow_index', 'psmrows', 'rownr')
        self.index_column('pepseq_index', 'peptide_sequences', 'sequence')
        self.index_column('pepid_index', 'peptide_sequences', 'pep_id')
        self.index_column('psmspepid_index', 'psms', 'pep_id')

    def index_protein_peptides(self):
        self.index_column('proteinpsm_index', 'protein_psm', 'protein_acc')
        self.index_column('protpsmid_index', 'protein_psm', 'psm_id')

    def get_exp_spectra_data_rows(self, shiftrows):
        cursor = self.get_cursor()
        rowlim = 'WHERE pr.rownr>{} '.format(shiftrows - 1) if shiftrows else ''
        return cursor.execute('SELECT pr.rownr, bs.set_name, sp.retention_time, '
                              'iit.ion_injection_time, im.ion_mobility '
                              'FROM psmrows AS pr '
                              'JOIN psms AS p USING(psm_id) '
                              'JOIN mzml AS sp USING(spectra_id) '
                              'LEFT OUTER JOIN ioninjtime AS iit USING(spectra_id) '
                              'LEFT OUTER JOIN ionmob AS im USING(spectra_id) '
                              'JOIN mzmlfiles as mf USING(mzmlfile_id) '
                              'JOIN biosets AS bs USING(set_id) ' + rowlim +
                              'ORDER BY pr.rownr')

    def drop_psm_indices(self):
        cursor = self.get_cursor()
        for index in ['psmid_index', 'psmspecid_index', 'psmrowid_index',
                'psmrow_index', 'pepseq_index', 'pepid_index', 'psmspepid_index',
                'proteinpsm_index', 'protpsmid_index']:
            cursor.execute('DROP INDEX IF EXISTS {}'.format(index))
        self.conn.commit()

    def drop_pgroup_tables(self):
        for table in ['protein_coverage',
                'protein_group_content',
                'psm_protein_groups',
                'protein_group_master']:
            self.conn.execute('DROP TABLE IF EXISTS {}'.format(table))
        self.conn.commit()

    def store_masters(self, allmasters, psm_masters):
        protids = self.get_protids()
        allmasters = ((protids[x],) for x in allmasters)
        cursor = self.get_cursor()
        cursor.executemany(
            'INSERT INTO protein_group_master(pacc_id) VALUES(?)',
            allmasters)
        master_ids = self.get_master_ids()
        psms = ((psm_id, master_ids[protids[master]])
                for psm_id, masters in psm_masters.items()
                for master in masters)
        cursor.executemany(
            'INSERT INTO psm_protein_groups(psm_id, master_id) '
            'VALUES(?, ?)', psms)
        self.conn.commit()
        self.index_column('psm_pg_index', 'psm_protein_groups', 'master_id')
        self.index_column('master_pacc_ix', 'protein_group_master', 'pacc_id')
        self.index_column('psm_pg_psmid_index', 'psm_protein_groups', 'psm_id')

    def update_master_proteins(self, new_masters):
        protids = self.get_protids()
        new_masters = ((protids[x[0]], x[1]) for x in new_masters)
        cur = self.get_cursor()
        sql = 'UPDATE protein_group_master SET pacc_id=? WHERE master_id=?'
        cur.executemany(sql, new_masters)
        self.conn.commit()

    def get_master_ids(self):
        cur = self.get_cursor()
        cur.execute('SELECT pacc_id, master_id FROM protein_group_master')
        return {p_acc: master_id for (p_acc, master_id) in cur}

    def store_coverage(self, coverage):
        cursor = self.get_cursor()
        sql = ('INSERT INTO protein_coverage(protein_acc, coverage) '
               'VALUES(?, ?)')
        cursor.executemany(sql, coverage)
        self.conn.commit()
        self.index_column('cov_index', 'protein_coverage', 'protein_acc')

    def store_protein_group_content(self, protein_groups):
        cursor = self.get_cursor()
        cursor.executemany('INSERT INTO protein_group_content('
                           'protein_acc, master_id, peptide_count, '
                           'psm_count, protein_score) '
                           'VALUES(?, ?, ?, ?, ?)', protein_groups)
        self.conn.commit()

    def index_protein_group_content(self):
        self.index_column('pgc_master_index', 'protein_group_content',
                          'master_id')

    def get_all_psm_protein_relations(self):
        sql = 'SELECT psm_id, protein_acc FROM protein_psm'
        cursor = self.get_cursor()
        return cursor.execute(sql)

    def get_proteins_for_peptide(self, psm_id):
        """Returns list of proteins for a passed psm_id"""
        protsql = self.get_sql_select(['protein_acc'], 'protein_psm')
        protsql = '{0} WHERE psm_id=?'.format(protsql)
        cursor = self.get_cursor()
        proteins = cursor.execute(protsql, psm_id).fetchall()
        return [x[0] for x in proteins]

    def get_psms_for_proteins(self, proteins):
        sql = """SELECT protein_acc, psm_id FROM protein_psm 
        WHERE protein_acc IN ({})""".format(','.join('?' * len(proteins)))
        return self.conn.execute(sql, proteins)

    def get_all_proteins_psms_seq(self):
        sql = ('SELECT p.protein_acc, ps.sequence, pp.psm_id, peps.sequence '
               'FROM proteins AS p '
               'JOIN protein_seq AS ps USING(protein_acc) '
               'JOIN protein_psm AS pp USING(protein_acc) '
               'JOIN psms AS psms USING(psm_id) '
               'JOIN peptide_sequences AS peps USING(pep_id)'
               )
        cursor = self.get_cursor()
        return cursor.execute(sql)

    def get_protein_psm_records(self):
        sql = ('SELECT protein_acc, psm_id FROM protein_psm')
        cursor = self.get_cursor()
        return cursor.execute(sql)

    def get_protein_group_candidates(self):
        sql = ('SELECT pgm.master_id, pgm.psm_id, pp.protein_acc, '
               'peps.sequence, p.score, pev.evidence_lvl, pc.coverage '
               'FROM psm_protein_groups AS pgm '
               'JOIN protein_psm AS pp USING(psm_id) '
               'JOIN psms AS p USING(psm_id) '
               'JOIN peptide_sequences AS peps USING(pep_id) '
               'LEFT OUTER JOIN protein_evidence AS pev '
               'ON pev.protein_acc=pp.protein_acc '
               'LEFT OUTER JOIN protein_coverage AS pc '
               'ON pc.protein_acc=pp.protein_acc '
               'ORDER BY pgm.master_id'
               )
        cursor = self.get_cursor()
        return cursor.execute(sql)

    def check_evidence_tables(self):
        """Returns True if there are records in evidence
        tables, otherwise returns False"""
        return self.check_table('protein_evidence') is not False

    def check_table(self, tablename):
        cursor = self.get_cursor()
        cursor.execute('SELECT * FROM {} LIMIT 10'.format(tablename))
        if len(cursor.fetchall()) > 0:
            return True
        return False

    def get_all_psms_proteingroups(self, evidence):
        fields = """
        SELECT pr.rownr, p.protein_acc, pgc.protein_acc, pgc.peptide_count,
        pgc.psm_count, pgc.protein_score, pc.coverage"""
        sql = """
        FROM psmrows AS pr
        JOIN psm_protein_groups AS ppg USING(psm_id)
        JOIN protein_group_master AS pgm USING(master_id)
        JOIN proteins AS p USING(pacc_id)
        JOIN protein_group_content AS pgc USING(master_id)
        JOIN protein_coverage AS pc ON pgc.protein_acc=pc.protein_acc
        """
        if evidence:
            sql = '{} {}'.format(sql, 
                    'JOIN protein_evidence AS pev ON pgc.protein_acc=pev.protein_acc')
            fields = '{}, pev.evidence_lvl '.format(fields)
        sql = '{} {} ORDER BY pr.rownr'.format(fields, sql)
        cursor = self.get_cursor()
        return cursor.execute(sql)

    def select_all_psm_quants(self, shiftrows, isobaric=False, precursor=False):
        selects = ['pr.rownr']
        joins = ['JOIN psms USING(psm_id)', 'JOIN mzml USING(spectra_id)']
        sqlfields, fieldcount = {}, 1
        if isobaric:
            selects.extend(['ic.channel_name', 'iq.intensity', 'pif.pif'])
            joins.extend(['JOIN isobaric_quant AS iq USING(spectra_id)',
                          'JOIN isobaric_channels AS ic USING(channel_id)',
                          'LEFT OUTER JOIN precursor_ion_fraction AS pif USING(spectra_id)'])
            sqlfields['isochan'] = fieldcount
            sqlfields['isoquant'] = fieldcount + 1
            sqlfields['pif'] = fieldcount + 2
            fieldcount += 3
        if precursor:
            selects.extend(['pq.intensity', 'pfw.fwhm'])
            joins.extend(['LEFT OUTER JOIN ms1_align USING(spectra_id)',
                          'LEFT OUTER JOIN ms1_quant AS pq USING(feature_id)',
                          'LEFT OUTER JOIN ms1_fwhm AS pfw USING(feature_id)',
                          ])
            sqlfields['precursor'] = fieldcount
            sqlfields['fwhm'] = fieldcount + 1

        rowlim = 'WHERE pr.rownr>{} '.format(shiftrows - 1) if shiftrows else ''
        sql = ('SELECT {} FROM psmrows as pr {} {} '
               'ORDER BY pr.rownr'.format(', '.join(selects), ' '.join(joins), rowlim))
        cursor = self.get_cursor()
        return cursor.execute(sql), sqlfields

    def get_all_quantmaps(self):
        """Returns all unique quant channels from lookup as list"""
        cursor = self.get_cursor()
        cursor.execute(
            'SELECT channel_name FROM isobaric_channels')
        return cursor.fetchall()

    def delete_sample_set_shift_rows(self, setnames):
        cursor = self.get_cursor()
        cursor.executemany('DELETE FROM biosets WHERE set_name=?', ((x,) for x in setnames))
        # Now all the rows will be gone where this set was, so we re-number:
        # Vacuuming updates the internal rowid column of the tables, which is
        # a count, when they are not INTEGER PRIMARY KEY
        self.conn.commit()
        self.conn.execute('DROP INDEX psmrowid_index')
        self.conn.execute('DROP INDEX psmrow_index')
        self.conn.execute('VACUUM')
        self.conn.execute('UPDATE psmrows SET rownr=rowid-1') # -1 since we start at 0
        self.conn.commit()
        # Put index back
        self.index_column('psmrowid_index', 'psmrows', 'psm_id')
        self.index_column('psmrow_index', 'psmrows', 'rownr')
        cursor = self.get_cursor()
        cursor.execute('DELETE FROM peptide_sequences WHERE pep_id IN ('
            'SELECT ps.pep_id FROM peptide_sequences AS ps '
            'LEFT OUTER JOIN psms ON ps.pep_id=psms.pep_id WHERE psms.pep_id IS NULL)')
        self.conn.commit()

    def get_mods_luciphor(self, ptms):
    # Fixed mods, var mods, ptms - {'mass': float, 'aa': str,  'name': str, 'pos': str}
        fix, var, ptm = [], [], []
        cursor = self.get_cursor()
        sql  = ('SELECT m.name, mp.aa, mp.variable, mp.pos, m.mass, mp.masschange '
                'FROM modpos AS mp JOIN mods AS m USING(mod_id)')
        for name, aa, var, pos, mass, newmass in cursor.execute(sql).fetchall():
            mod = {'mass': mass, 'masschange': masschange, 'aa': aa, 'name': name, 'pos': pos}
            if not var:
                # fixed mod
                fix.append(mod)
            elif name in ptms and (aa in ptms[name] or ptms[name] == ''):
                ptm.append(mod)
            else:
                var.append(mod)
        return fix, var, ptm



def write_tsv(headerfields, features, outfn):
    """Writes header and generator of lines to tab separated file.

    headerfields - list of field names in header in correct order
    features - generates 1 list per line that belong to header
    outfn - filename to output to. Overwritten if exists
    """
    with open(outfn, 'w') as fp:
        write_tsv_line_from_list(headerfields, fp)
        for line in features:
            write_tsv_line_from_list([str(line[field]) for field
                                      in headerfields], fp)


def write_multi_mzidtsv(header, oldheader, psms, base_outfile):
    outfile_handles = {}
    for psm in psms:
        line = psm['psm']
        split_pool = psm['split_pool']
        if split_pool not in outfile_handles:
            outfile_handles[split_pool] = open(base_outfile.format(split_pool),
                                               'w')
            write_tsv_line_from_list(header, outfile_handles[split_pool])
        write_tsv_line_from_list([line[field] for field in oldheader],
                                     outfile_handles[split_pool])
    [handle.close() for handle in outfile_handles.values()]


def write_tsv_line_from_list(linelist, outfp):
    """Utility method to convert list to tsv line with carriage return"""
    line = '\t'.join(linelist)
    outfp.write(line)
    outfp.write('\n')


def parse_NA(feature, header):
    for field in header:
        try:
            feature[field] = str(feature[field])
        except KeyError:
            feature[field] = 'NA'
    return feature


def write_table_with_na(header, features, outfn):
    na_feats = (parse_NA(x, header) for x in features)
    write_tsv(header, na_feats, outfn)



import re

from app.lookups.sqlite.protpeptable import ProtPepTable
from app.dataformats import peptable as ph
from app.dataformats import prottable as proth


class PepTableProteinCentricDB(ProtPepTable):
    datatype = 'peptide'
    stdheaderfields = [
            ph.HEADER_NO_PSM,
            ph.HEADER_QVAL,
            ph.HEADER_PEP,
            ]
    singlefields = stdheaderfields + [
            ph.HEADER_FALSE_LOC_RATE,
            ph.HEADER_AREA,
            proth.HEADER_NO_FULLQ_PSMS,
            ]
    colmap = {'peptide_sequences': ['pep_id', 'sequence'],
              'peptide_precur_quanted': ['pep_id', 'peptable_id', 'quant'],
              'peptide_fdr': ['pep_id', 'peptable_id', 'fdr'],
              'peptide_pep': ['pep_id', 'peptable_id', 'pep'],
              'ptm_flr': ['pep_id', 'peptable_id', 'flr'],
              'peptide_iso_fullpsms': ['pep_id', 'peptable_id', 'amount_psms'],
              'pepquant_channels': ['channel_id', 'peptable_id',
                                    'channel_name', 'amount_psms_name'],
              'peptide_iso_quanted': ['peptidequant_id', 'pep_id',
                                      'channel_id', 'quantvalue',
                                      'amount_psms'],
              }

    def add_tables(self, tabletypes=[]):
        ttypes = ['fntable', 'isoqtable', 'isochtable', 'prectable',
                'fdrtable', 'peptable', 'fullqpsmtable', 'flrtable']
        self.create_tables([self.table_map[self.datatype][x] for x in ttypes
            if x in self.table_map[self.datatype]])

    def store_pep(self, pep):
        self.store_singlecol('peptable', pep)
        table = self.table_map[self.datatype]['peptable']
        self.singlecols_to_index.append(('pep_acc_ix', table, self.colmap[table][0]))
        self.singlecols_to_index.append(('pep_table_ix', table, self.colmap[table][1]))

    def create_pdata_map(self):
        """This runs only once, returns the data which is not dependent on sets,
        in a dict with accessions as keys"""
        # First protein SQL, some peptides have TOO MANY proteingroup matches 
        # for SQLite to output on a single line, therefore we loop rows 
        # instead of doing GROUP_CONCAT
        protsql = """
        SELECT sub.pep_id, sub.seq, sub.pseq, sub.pacc, sub.pgcnr, IFNULL(pd.description, 'NA'), pc.coverage
        FROM (
            SELECT psms.pep_id AS pep_id, ps.sequence AS seq, pseq.sequence AS pseq, pgm.master_id,
                pgm.pacc_id AS pacc_id, p.protein_acc AS pacc, COUNT(pgc.protein_acc) AS pgcnr
            FROM psm_protein_groups AS pp
            INNER JOIN psms ON psms.psm_id=pp.psm_id
            INNER JOIN protein_group_master AS pgm ON pgm.master_id=pp.master_id
            INNER JOIN proteins AS p ON p.pacc_id=pgm.pacc_id
            INNER JOIN protein_seq AS pseq ON p.protein_acc=pseq.protein_acc
            INNER JOIN protein_group_content AS pgc ON pgm.master_id=pgc.master_id
            INNER JOIN peptide_sequences AS ps ON psms.pep_id=ps.pep_id
            GROUP BY psms.pep_id, pgm.master_id
            ) AS sub
            LEFT OUTER JOIN protein_coverage AS pc ON sub.pacc=pc.protein_acc
            LEFT OUTER JOIN prot_desc AS pd ON sub.pacc_id=pd.pacc_id
            """
        pgdata = {}
        cursor = self.get_cursor()
        for pid, seq, pseq, pacc, pgroupnr, desc, cov in cursor.execute(protsql):
            barepep = re.sub('[^A-Za-z]', '', seq)
            start = pseq.index(barepep) + 1
            stop = start + len(barepep) - 1
            if pid in pgdata:
                pgdata[pid][ph.HEADER_PROTEINS].append(pacc)
                pgdata[pid][ph.HEADER_STARTSTOP].append(f'{start}-{stop}')
                pgdata[pid][ph.HEADER_NO_CONTENTPROTEINS].append(str(pgroupnr))
                pgdata[pid][ph.HEADER_DESCRIPTIONS].append(desc)
                pgdata[pid][ph.HEADER_COVERAGES].append(str(cov))
            else:
                pgdata[pid] = {
                    ph.HEADER_PROTEINS: [pacc],
                    ph.HEADER_BAREPEP: [barepep],
                    ph.HEADER_STARTSTOP: [f'{start}-{stop}'],
                    ph.HEADER_NO_CONTENTPROTEINS: [str(pgroupnr)],
                    ph.HEADER_DESCRIPTIONS: [desc],
                    ph.HEADER_COVERAGES: [str(cov)],
                    ph.HEADER_GENES:  [],
                    ph.HEADER_ASSOCIATED: [],
                    }

        # Now get the genes
        ensg_sql = """
            SELECT DISTINCT psms.pep_id, IFNULL(g.gene_acc, 'NA')
            FROM psms
            INNER JOIN psm_protein_groups AS ppg ON psms.psm_id=ppg.psm_id
            INNER JOIN protein_group_master AS pgm ON pgm.master_id=ppg.master_id
            LEFT OUTER JOIN ensg_proteins AS egp ON egp.pacc_id=pgm.pacc_id
            LEFT OUTER JOIN genes AS g ON g.gene_id=egp.gene_id
            """

        gn_sql = """
            SELECT DISTINCT psms.pep_id, IFNULL(aid.assoc_id, 'NA')
            FROM psms
            INNER JOIN psm_protein_groups AS ppg ON psms.psm_id=ppg.psm_id
            INNER JOIN protein_group_master AS pgm ON pgm.master_id=ppg.master_id
            LEFT OUTER JOIN genename_proteins AS gnp ON gnp.pacc_id=pgm.pacc_id
            LEFT OUTER JOIN associated_ids AS aid ON aid.gn_id=gnp.gn_id
            """
        for pid, ensg in cursor.execute(ensg_sql):
            pgdata[pid][ph.HEADER_GENES].append(ensg)
        for pid, gene in cursor.execute(gn_sql):
            pgdata[pid][ph.HEADER_ASSOCIATED].append(gene)

        # Finish with creating strings of the lists
        for pid, headers in pgdata.items():
            for header, vals in headers.items():
                pgdata[pid][header] = ';'.join(vals)

        pepsql = 'SELECT pep_id, sequence FROM peptide_sequences'
        cursor = self.get_cursor()
        for pid, seq in cursor.execute(pepsql):
            pgdata[pid][ph.HEADER_PEPTIDE] = seq
        return pgdata

    def merge_features(self):
        sql = """
    SELECT bs.set_name, ps.pep_id, COUNT(DISTINCT psms.psm_id), pf.fdr, pp.pep,
            flr.flr, ppq.quant, fqpsm.amount_psms, GROUP_CONCAT(pqc.channel_name),
            GROUP_CONCAT(piq.quantvalue), GROUP_CONCAT(piq.amount_psms)
        FROM peptide_sequences AS ps
        INNER JOIN psms ON ps.pep_id=psms.pep_id 
        INNER JOIN mzml ON psms.spectra_id=mzml.spectra_id
        INNER JOIN mzmlfiles ON mzml.mzmlfile_id=mzmlfiles.mzmlfile_id
        INNER JOIN biosets AS bs ON mzmlfiles.set_id=bs.set_id
        INNER JOIN peptide_tables AS pt ON pt.set_id=bs.set_id
        INNER JOIN peptide_fdr AS pf ON pf.peptable_id=pt.peptable_id AND 
            pf.pep_id=ps.pep_id
        LEFT OUTER JOIN peptide_pep AS pp ON pp.peptable_id=pt.peptable_id AND 
            pp.pep_id=ps.pep_id
        LEFT OUTER JOIN ptm_flr AS flr ON flr.peptable_id=pt.peptable_id AND
            flr.pep_id=ps.pep_id
        LEFT OUTER JOIN peptide_precur_quanted AS ppq ON ppq.peptable_id=pt.peptable_id AND
            ppq.pep_id=ps.pep_id
        LEFT OUTER JOIN peptide_iso_fullpsms AS fqpsm ON fqpsm.peptable_id=pt.peptable_id AND
            fqpsm.pep_id=ps.pep_id
        LEFT OUTER JOIN pepquant_channels AS pqc ON pqc.peptable_id=pt.peptable_id
        LEFT OUTER JOIN peptide_iso_quanted AS piq ON piq.channel_id=pqc.channel_id AND
            piq.pep_id=ps.pep_id
        GROUP BY ps.pep_id, bs.set_id
        """
        cursor = self.get_cursor()
        cursor.execute(sql)
        return cursor



class PepTableGeneCentricDB(PepTableProteinCentricDB):
    datatype = 'peptide'

    def create_pdata_map(self):
        """This runs only once, returns the data which is not dependent on sets,
        in a dict with accessions as keys. Different from protein grouped
        data in that we dont pull proteins from the protein group master table.
        """

        sql = """
SELECT ps.pep_id, ps.sequence, GROUP_CONCAT(IFNULL(gsub.ensg, 'NA'), ';'),
    GROUP_CONCAT(IFNULL(gnsub.gn, 'NA'), ';')
    FROM peptide_sequences AS ps
    LEFT OUTER JOIN (
        SELECT DISTINCT psms.pep_id AS pid, g.gene_acc AS ensg
        FROM genes as g
        INNER JOIN ensg_proteins AS egp ON egp.gene_id=g.gene_id
        INNER JOIN proteins AS p ON p.pacc_id=egp.pacc_id
        INNER JOIN protein_psm AS pp ON pp.protein_acc=p.protein_acc
        INNER JOIN psms ON psms.psm_id=pp.psm_id
        ) AS gsub ON gsub.pid=ps.pep_id
    LEFT OUTER JOIN (
        SELECT DISTINCT psms.pep_id AS pid, aid.assoc_id AS gn
        FROM associated_ids AS aid
        INNER JOIN genename_proteins AS gnp ON gnp.gn_id=aid.gn_id
        INNER JOIN proteins AS p ON p.pacc_id=gnp.pacc_id
        INNER JOIN protein_psm AS pp ON pp.protein_acc=p.protein_acc
        INNER JOIN psms ON psms.psm_id=pp.psm_id
        ) AS gnsub ON gnsub.pid=ps.pep_id
    GROUP BY ps.pep_id
    """
        cursor = self.get_cursor()
        pgdata = {}
        for pid, seq, gaccs, aids in cursor.execute(sql):
            barepep = re.sub('[^A-Za-z]', '', seq)
            pgdata[pid] = {
                    ph.HEADER_PEPTIDE: seq,
                    ph.HEADER_BAREPEP: barepep,
                    ph.HEADER_GENES: gaccs,
                    ph.HEADER_ASSOCIATED: aids,
                    }
        return pgdata

    def merge_features(self):
        sql = """
    SELECT bs.set_name, ps.pep_id, COUNT(DISTINCT psms.psm_id), pf.fdr, pp.pep,
            flr.flr, ppq.quant, fqpsm.amount_psms, GROUP_CONCAT(pqc.channel_name),
            GROUP_CONCAT(piq.quantvalue), GROUP_CONCAT(piq.amount_psms)
        FROM peptide_sequences AS ps
        JOIN biosets AS bs 
        INNER JOIN psms ON ps.pep_id=psms.pep_id 
        INNER JOIN peptide_tables AS pt ON pt.set_id=bs.set_id
        INNER JOIN peptide_fdr AS pf ON pf.peptable_id=pt.peptable_id AND 
            pf.pep_id=ps.pep_id
        LEFT OUTER JOIN peptide_pep AS pp ON pp.peptable_id=pt.peptable_id AND 
            pp.pep_id=ps.pep_id
        LEFT OUTER JOIN ptm_flr AS flr ON flr.peptable_id=pt.peptable_id AND
            flr.pep_id=ps.pep_id
        LEFT OUTER JOIN peptide_precur_quanted AS ppq ON ppq.peptable_id=pt.peptable_id AND
            ppq.pep_id=ps.pep_id
        LEFT OUTER JOIN peptide_iso_fullpsms AS fqpsm ON fqpsm.peptable_id=pt.peptable_id AND
            fqpsm.pep_id=ps.pep_id
        LEFT OUTER JOIN pepquant_channels AS pqc ON pqc.peptable_id=pt.peptable_id
        LEFT OUTER JOIN peptide_iso_quanted AS piq ON piq.channel_id=pqc.channel_id AND
            piq.pep_id=ps.pep_id
        GROUP BY ps.pep_id, bs.set_id
        """
        cursor = self.get_cursor()
        cursor.execute(sql)
        return cursor



class PepTablePlainDB(PepTableProteinCentricDB):
    datatype = 'peptide'

    def create_pdata_map(self):
        """This runs only once, returns the data which is not dependent on sets,
        in a dict with accessions as keys. 
        In plain DB we only output peptides, not proteins etc
        """
        # FIXME LEFT OUTER JOIN protein_seq is normally INNER JOIN, but there is an
        # edge case since we use it in dda pipeline for PTM merging. If PTMs are ever
        # included in msstitch, we could fix this.
        sql = """
    SELECT ps.pep_id, ps.sequence, GROUP_CONCAT(ppp.protein_acc, ';'), GROUP_CONCAT(ppp.sequence, ';')
    FROM peptide_sequences AS ps
    INNER JOIN (
            SELECT DISTINCT psms.pep_id, pp.protein_acc, pseq.sequence
            FROM psms
            INNER JOIN protein_psm AS pp ON pp.psm_id=psms.psm_id
            LEFT OUTER JOIN protein_seq AS pseq ON pp.protein_acc=pseq.protein_acc
            ) AS ppp ON ppp.pep_id=ps.pep_id
    GROUP BY ps.pep_id
        """
        cursor = self.get_cursor()
        pgdata = {}
        for pid, seq, prots, pseqs in cursor.execute(sql):
            barepep = re.sub('[^A-Za-z]', '', seq)
            startstop = []
            pseqs_or_na = pseqs.split(';') if pseqs else []
            for pseq in pseqs_or_na:
                start = pseq.index(barepep) + 1
                stop = start + len(barepep) - 1
                startstop.append(f'{start}-{stop}')
            if pid in pgdata:
                pgdata[pid][ph.HEADER_PROTEINS].append(pacc)
                pgdata[pid][ph.HEADER_STARTSTOP].append(f'{start}-{stop}')
            pgdata[pid] = {
                    ph.HEADER_PEPTIDE: seq,
                    ph.HEADER_BAREPEP: barepep,
                    ph.HEADER_PROTEINS: prots,
                    ph.HEADER_STARTSTOP: ';'.join(startstop)
                    }
        return pgdata

    def merge_features(self):
        sql = """
    SELECT bs.set_name, ps.pep_id, COUNT(DISTINCT psms.psm_id), pf.fdr, pp.pep,
            flr.flr, ppq.quant, fqpsm.amount_psms, GROUP_CONCAT(pqc.channel_name),
            GROUP_CONCAT(piq.quantvalue), GROUP_CONCAT(piq.amount_psms)
        FROM peptide_sequences AS ps
        JOIN biosets AS bs 
        INNER JOIN psms ON ps.pep_id=psms.pep_id 
        INNER JOIN peptide_tables AS pt ON pt.set_id=bs.set_id
        INNER JOIN peptide_fdr AS pf ON pf.peptable_id=pt.peptable_id AND 
            pf.pep_id=ps.pep_id
        LEFT OUTER JOIN peptide_pep AS pp ON pp.peptable_id=pt.peptable_id AND 
            pp.pep_id=ps.pep_id
        LEFT OUTER JOIN ptm_flr AS flr ON flr.peptable_id=pt.peptable_id AND flr.pep_id=ps.pep_id
        LEFT OUTER JOIN peptide_precur_quanted AS ppq ON ppq.peptable_id=pt.peptable_id AND
            ppq.pep_id=ps.pep_id
        LEFT OUTER JOIN peptide_iso_fullpsms AS fqpsm ON fqpsm.peptable_id=pt.peptable_id AND
            fqpsm.pep_id=ps.pep_id
        LEFT OUTER JOIN pepquant_channels AS pqc ON pqc.peptable_id=pt.peptable_id
        LEFT OUTER JOIN peptide_iso_quanted AS piq ON piq.channel_id=pqc.channel_id AND
            piq.pep_id=ps.pep_id
        GROUP BY ps.pep_id, bs.set_id
        """
        cursor = self.get_cursor()
        cursor.execute(sql)
        return cursor
